"""Item-level tests for the in-memory ChatKit store."""

from __future__ import annotations
from datetime import UTC, datetime
import pytest
from chatkit.types import InferenceOptions, UserMessageItem, UserMessageTextContent
from orcheo_backend.app.chatkit import (
    ChatKitRequestContext,
    InMemoryChatKitStore,
)


def _build_user_item(thread_id: str, *, index: int) -> UserMessageItem:
    return UserMessageItem(
        id=f"msg_{index}",
        thread_id=thread_id,
        created_at=datetime(2024, 1, 1, hour=index, tzinfo=UTC),
        content=[UserMessageTextContent(type="input_text", text=f"Message {index}")],
        attachments=[],
        quoted_text=None,
        inference_options=InferenceOptions(),
    )


@pytest.mark.asyncio
async def test_in_memory_store_load_thread_items_pagination() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_items"

    items = [_build_user_item(thread_id, index=i) for i in range(4)]
    for item in items:
        await store.add_thread_item(thread_id, item, context)

    page1 = await store.load_thread_items(
        thread_id, after=None, limit=2, order="asc", context=context
    )
    assert len(page1.data) == 2
    assert page1.has_more is True

    page2 = await store.load_thread_items(
        thread_id, after=page1.data[-1].id, limit=2, order="asc", context=context
    )
    assert len(page2.data) == 2
    assert page2.has_more is False


@pytest.mark.asyncio
async def test_in_memory_store_load_thread_items_descending() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_desc"

    items = [_build_user_item(thread_id, index=i) for i in range(3)]
    for item in items:
        await store.add_thread_item(thread_id, item, context)

    page = await store.load_thread_items(
        thread_id, after=None, limit=10, order="desc", context=context
    )
    assert page.data[0].id == "msg_2"
    assert page.data[-1].id == "msg_0"


@pytest.mark.asyncio
async def test_in_memory_store_save_item() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_save"

    item = _build_user_item(thread_id, index=1)
    await store.save_item(thread_id, item, context)

    loaded = await store.load_item(thread_id, "msg_1", context)
    assert loaded.content[0].text == "Message 1"

    updated_item = UserMessageItem(
        id="msg_1",
        thread_id=thread_id,
        created_at=datetime.now(UTC),
        content=[UserMessageTextContent(type="input_text", text="Updated")],
        attachments=[],
        quoted_text=None,
        inference_options=InferenceOptions(),
    )
    await store.save_item(thread_id, updated_item, context)

    loaded_updated = await store.load_item(thread_id, "msg_1", context)
    assert loaded_updated.content[0].text == "Updated"


@pytest.mark.asyncio
async def test_in_memory_store_load_item_not_found() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}

    from chatkit.store import NotFoundError

    with pytest.raises(NotFoundError):
        await store.load_item("thr_missing", "msg_missing", context)


@pytest.mark.asyncio
async def test_in_memory_store_delete_thread_item() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_delete"

    item = _build_user_item(thread_id, index=0)
    await store.add_thread_item(thread_id, item, context)

    await store.delete_thread_item(thread_id, item.id, context)

    page = await store.load_thread_items(
        thread_id, after=None, limit=10, order="asc", context=context
    )
    assert len(page.data) == 0


@pytest.mark.asyncio
async def test_in_memory_store_save_item_iterates_through_non_matching() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_iter"

    for i in range(3):
        await store.add_thread_item(
            thread_id,
            _build_user_item(thread_id, index=i),
            context,
        )

    new_item = UserMessageItem(
        id="msg_new",
        thread_id=thread_id,
        created_at=datetime.now(UTC),
        content=[UserMessageTextContent(type="input_text", text="New message")],
        attachments=[],
        quoted_text=None,
        inference_options=InferenceOptions(),
    )
    await store.save_item(thread_id, new_item, context)

    loaded = await store.load_item(thread_id, "msg_new", context)
    assert loaded.content[0].text == "New message"


@pytest.mark.asyncio
async def test_in_memory_store_load_item_iterates_through_non_matching() -> None:
    store = InMemoryChatKitStore()
    context: ChatKitRequestContext = {}
    thread_id = "thr_search"

    for i in range(3):
        await store.add_thread_item(
            thread_id,
            _build_user_item(thread_id, index=i),
            context,
        )

    from chatkit.store import NotFoundError

    with pytest.raises(NotFoundError):
        await store.load_item(thread_id, "msg_nonexistent", context)



"""Unit tests for the RunHistoryRecord dataclass behavior."""

from __future__ import annotations
from orcheo_backend.app.history import RunHistoryRecord


def test_run_history_record_mark_failed_sets_error() -> None:
    record = RunHistoryRecord(workflow_id="wf", execution_id="exec")
    record.mark_failed("boom")

    assert record.status == "error"
    assert record.error == "boom"
    assert record.completed_at is not None


def test_run_history_record_mark_cancelled_sets_status() -> None:
    record = RunHistoryRecord(workflow_id="wf", execution_id="exec")
    record.mark_cancelled(reason="shutdown")

    assert record.status == "cancelled"
    assert record.error == "shutdown"
    assert record.completed_at is not None


def test_run_history_record_append_step_increments_index() -> None:
    record = RunHistoryRecord(workflow_id="wf", execution_id="exec")
    step1 = record.append_step({"action": "start"})
    step2 = record.append_step({"action": "continue"})

    assert step1.index == 0
    assert step2.index == 1
    assert len(record.steps) == 2
    assert record.steps[0].payload == {"action": "start"}
    assert record.steps[1].payload == {"action": "continue"}


def test_run_history_record_mark_completed_clears_error() -> None:
    record = RunHistoryRecord(workflow_id="wf", execution_id="exec")
    record.error = "previous error"
    record.mark_completed()

    assert record.status == "completed"
    assert record.error is None
    assert record.completed_at is not None



"""Web scraping and MongoDB vector upload workflow.

Scrapes given web pages, chunks the body text, generates vector embeddings,
and uploads the results to a MongoDB collection with source URL metadata.

Configurable inputs (config.json):
- database: MongoDB database name
- collection: MongoDB collection name
- vector_path: Document field used for the embedding vector
- embedding_method: Registered embedding method identifier
"""

from langgraph.graph import END, START, StateGraph
from orcheo.graph.state import State
from orcheo.nodes.conversational_search import (
    OPENAI_TEXT_EMBEDDING_3_SMALL,
    ChunkEmbeddingNode,
    ChunkingStrategyNode,
    WebDocumentLoaderNode,
)
from orcheo.nodes.mongodb import MongoDBInsertManyNode


async def orcheo_workflow() -> StateGraph:
    """Build the web scraping and MongoDB upload workflow."""
    web_loader = WebDocumentLoaderNode(
        name="web_loader",
        urls="{{config.configurable.urls}}",
    )

    chunking = ChunkingStrategyNode(
        name="chunking",
        source_result_key="web_loader",
        chunk_size=800,
        chunk_overlap=80,
    )

    chunk_embedding = ChunkEmbeddingNode(
        name="chunk_embedding",
        source_result_key="chunking",
        embedding_methods={"dense": OPENAI_TEXT_EMBEDDING_3_SMALL},
    )

    mongodb_upload = MongoDBInsertManyNode(
        name="mongodb_upload",
        database="{{config.configurable.database}}",
        collection="{{config.configurable.collection}}",
        source_result_key="chunk_embedding",
        embedding_name="dense",
        vector_field="{{config.configurable.vector_path}}",
        text_field="body",
    )

    graph = StateGraph(State)
    graph.add_node("web_loader", web_loader)
    graph.add_node("chunking", chunking)
    graph.add_node("chunk_embedding", chunk_embedding)
    graph.add_node("mongodb_upload", mongodb_upload)
    graph.add_edge(START, "web_loader")
    graph.add_edge("web_loader", "chunking")
    graph.add_edge("chunking", "chunk_embedding")
    graph.add_edge("chunk_embedding", "mongodb_upload")
    graph.add_edge("mongodb_upload", END)

    return graph



"""Credential and workflow coverage tests for orcheo_backend.app."""

from __future__ import annotations
import importlib
from typing import Any
from uuid import uuid4
import pytest
from fastapi.testclient import TestClient


backend_app = importlib.import_module("orcheo_backend.app")


@pytest.mark.asyncio
async def test_credential_health_get_without_service(
    api_client: TestClient, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Credential health endpoint returns 503 when service missing."""
    workflow_response = api_client.post(
        "/api/workflows",
        json={"name": "Test Flow", "actor": "tester"},
    )
    workflow_id = workflow_response.json()["id"]

    monkeypatch.setitem(backend_app._credential_service_ref, "service", None)
    api_client.app.dependency_overrides[backend_app.get_credential_service] = (
        lambda: None
    )

    response = api_client.get(f"/api/workflows/{workflow_id}/credentials/health")
    assert response.status_code == 503
    assert "not configured" in response.json()["detail"]


@pytest.mark.asyncio
async def test_credential_health_validate_without_service(
    api_client: TestClient, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Credential validation endpoint returns 503 when service missing."""
    workflow_response = api_client.post(
        "/api/workflows",
        json={"name": "Test Flow", "actor": "tester"},
    )
    workflow_id = workflow_response.json()["id"]

    monkeypatch.setitem(backend_app._credential_service_ref, "service", None)
    api_client.app.dependency_overrides[backend_app.get_credential_service] = (
        lambda: None
    )

    response = api_client.post(
        f"/api/workflows/{workflow_id}/credentials/validate",
        json={"actor": "tester"},
    )
    assert response.status_code == 503
    assert "not configured" in response.json()["detail"]


def test_delete_credential_not_found(api_client: TestClient) -> None:
    """Deleting a non-existent credential returns 404."""
    missing_id = uuid4()
    response = api_client.delete(f"/api/credentials/{missing_id}")
    assert response.status_code == 404


def test_delete_credential_scope_violation(api_client: TestClient) -> None:
    """Deleting credential with mismatched workflow raises 403."""
    workflow_id = uuid4()
    other_workflow_id = uuid4()

    create_response = api_client.post(
        "/api/credentials",
        json={
            "name": "Scoped Cred",
            "provider": "test",
            "secret": "secret",
            "actor": "tester",
            "access": "private",
            "workflow_id": str(workflow_id),
        },
    )
    credential_id = create_response.json()["id"]

    response = api_client.delete(
        f"/api/credentials/{credential_id}",
        params={"workflow_id": str(other_workflow_id)},
    )
    assert response.status_code == 403


def test_create_credential_with_value_error(
    api_client: TestClient, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Backend surfaces ValueError from credential vault as 422."""
    vault = api_client.app.state.vault

    def mock_create_credential(*args: Any, **kwargs: Any) -> None:
        raise ValueError("Invalid credential configuration")

    monkeypatch.setattr(vault, "create_credential", mock_create_credential)

    response = api_client.post(
        "/api/credentials",
        json={
            "name": "Test Cred",
            "provider": "test",
            "secret": "secret",
            "actor": "tester",
            "access": "public",
            "kind": "secret",
        },
    )
    assert response.status_code == 422


def test_list_workflows_includes_archived(api_client: TestClient) -> None:
    """Workflows endpoint optionally returns archived entries."""
    create_response = api_client.post(
        "/api/workflows",
        json={"name": "To Archive", "actor": "tester"},
    )
    workflow_id = create_response.json()["id"]

    api_client.delete(f"/api/workflows/{workflow_id}", params={"actor": "tester"})

    response = api_client.get("/api/workflows")
    assert workflow_id not in [w["id"] for w in response.json()]

    response = api_client.get("/api/workflows?include_archived=true")
    assert any(w["id"] == workflow_id for w in response.json())


def test_ingest_workflow_version_script_error(api_client: TestClient) -> None:
    """Ingest endpoint returns 400 for bad script payloads."""
    workflow_response = api_client.post(
        "/api/workflows",
        json={"name": "Bad Script", "actor": "tester"},
    )
    workflow_id = workflow_response.json()["id"]

    response = api_client.post(
        f"/api/workflows/{workflow_id}/versions/ingest",
        json={
            "script": "invalid python code!!!",
            "entrypoint": "app",
            "created_by": "tester",
        },
    )
    assert response.status_code == 400



"""Tests covering repository and FastAPI app factory helpers."""

import importlib
import json
import sys
from types import ModuleType
import pytest
from orcheo_backend.app import _create_repository, create_app, get_repository
from orcheo_backend.app._app_module import _AppModule, install_app_module_proxy
from orcheo_backend.app.factory import _DEFAULT_ALLOWED_ORIGINS, _load_allowed_origins
from orcheo_backend.app.repository import InMemoryWorkflowRepository


backend_module = importlib.import_module("orcheo_backend.app")


def test_install_app_module_proxy_is_idempotent(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Installing the proxy twice should be a no-op on the second call."""
    module_name = "tests.dummy_app_module"
    dummy_module = ModuleType(module_name)
    monkeypatch.setitem(sys.modules, module_name, dummy_module)

    install_app_module_proxy(module_name)
    proxied_module = sys.modules[module_name]
    assert isinstance(proxied_module, _AppModule)

    install_app_module_proxy(module_name)
    assert sys.modules[module_name] is proxied_module

    sys.modules.pop(module_name, None)


def test_app_module_exposes_sensitive_debug_flag(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """The module proxy forwards _should_log_sensitive_debug lookups."""
    sentinel = object()
    monkeypatch.setattr(
        backend_module._workflow_execution_module,  # type: ignore[attr-defined]
        "_should_log_sensitive_debug",
        sentinel,
        raising=False,
    )

    monkeypatch.delattr(backend_module, "_should_log_sensitive_debug", raising=False)

    assert backend_module._should_log_sensitive_debug is sentinel  # type: ignore[attr-defined]


def test_get_repository_returns_singleton() -> None:
    """The module-level repository accessor returns a singleton instance."""

    first = get_repository()
    second = get_repository()
    assert first is second


def test_create_app_allows_dependency_override() -> None:
    """Passing a repository instance wires it into FastAPI dependency overrides."""

    repository = InMemoryWorkflowRepository()
    app = create_app(repository)

    override = app.dependency_overrides[get_repository]
    assert override() is repository


def test_create_repository_inmemory_backend(monkeypatch: pytest.MonkeyPatch) -> None:
    """The application factory instantiates the in-memory repository."""

    class DummySettings:
        repository_backend = "inmemory"
        repository_sqlite_path = "ignored.sqlite"

    monkeypatch.setattr(backend_module, "get_settings", lambda: DummySettings())

    repository = _create_repository()
    assert isinstance(repository, InMemoryWorkflowRepository)


def test_create_repository_invalid_backend(monkeypatch: pytest.MonkeyPatch) -> None:
    """Unsupported repository backends raise a clear error."""

    class DummySettings:
        repository_backend = "invalid_backend"
        repository_sqlite_path = "ignored.sqlite"

    monkeypatch.setattr(backend_module, "get_settings", lambda: DummySettings())

    with pytest.raises(ValueError, match="Unsupported repository backend"):
        _create_repository()


def test_load_allowed_origins_reads_json_list(monkeypatch: pytest.MonkeyPatch) -> None:
    """JSON arrays should be parsed, trimmed, and filtered."""
    monkeypatch.setenv(
        "ORCHEO_CORS_ALLOW_ORIGINS",
        json.dumps([" https://foo.example ", ""]),
    )

    origins = _load_allowed_origins()
    assert origins == ["https://foo.example"]


def test_load_allowed_origins_reads_csv_string(monkeypatch: pytest.MonkeyPatch) -> None:
    """Comma-separated values fallback when JSON parsing fails."""
    monkeypatch.setenv(
        "ORCHEO_CORS_ALLOW_ORIGINS",
        "https://a.example, ,https://b.example  ",
    )

    origins = _load_allowed_origins()
    assert origins == ["https://a.example", "https://b.example"]


def test_load_allowed_origins_defaults_when_empty(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """If the filtered list is empty the defaults should be returned."""
    monkeypatch.setenv(
        "ORCHEO_CORS_ALLOW_ORIGINS",
        json.dumps(["", "   "]),
    )

    origins = _load_allowed_origins()
    assert origins == list(_DEFAULT_ALLOWED_ORIGINS)
    assert origins is not _DEFAULT_ALLOWED_ORIGINS


def test_load_allowed_origins_defaults_when_unset(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """When the env var is unset, the defaults should be returned."""
    monkeypatch.delenv("ORCHEO_CORS_ALLOW_ORIGINS", raising=False)

    origins = _load_allowed_origins()
    assert origins == list(_DEFAULT_ALLOWED_ORIGINS)
    assert origins is not _DEFAULT_ALLOWED_ORIGINS



"""Public interface for the trigger orchestration layer."""

from orcheo.triggers.layer.core import TriggerLayer
from orcheo.triggers.layer.models import (
    CronDispatchPlan,
    ManualDispatchPlan,
    StateCleanupConfig,
    TriggerDispatch,
)


__all__ = [
    "CronDispatchPlan",
    "ManualDispatchPlan",
    "StateCleanupConfig",
    "TriggerDispatch",
    "TriggerLayer",
]



"""Cover additional branches in authentication websocket helpers."""

from __future__ import annotations
from types import SimpleNamespace
from orcheo_backend.app.authentication.dependencies import (
    _extract_websocket_protocol_token,
    _resolve_websocket_token,
)


def test_extract_websocket_protocol_token_returns_none_for_blank_header() -> None:
    """Empty or whitespace-only headers do not yield tokens or subprotocols."""

    assert _extract_websocket_protocol_token(None) == (None, None)
    assert _extract_websocket_protocol_token(" ,,   ") == (None, None)


def test_extract_websocket_protocol_token_finds_bearer_token_and_subprotocol() -> None:
    """Token parsing ignores empty bearer entries and reports the
    orcheo-auth subprotocol.
    """

    header = "foo, orcheo-auth, bearer., bearer.valid-token, bar"
    token, subprotocol = _extract_websocket_protocol_token(header)

    assert token == "valid-token"
    assert subprotocol == "orcheo-auth"


def test_resolve_websocket_token_prefers_authorization_header() -> None:
    """An explicit Authorization header takes precedence over websocket
    protocols.
    """

    headers = {
        "sec-websocket-protocol": "orcheo-auth",
        "authorization": "Bearer auth-header-token",
    }
    websocket = SimpleNamespace(headers=headers)

    token, subprotocol = _resolve_websocket_token(websocket)
    assert token == "auth-header-token"
    assert subprotocol == "orcheo-auth"


def test_resolve_websocket_token_uses_protocol_token_without_authorization() -> None:
    """Fall back to the protocol bearer token on Authorization header absent."""

    headers = {"sec-websocket-protocol": "bearer.protocol-token"}
    websocket = SimpleNamespace(headers=headers)

    token, subprotocol = _resolve_websocket_token(websocket)
    assert token == "protocol-token"
    assert subprotocol is None



"""ChatKit-related request/response schemas."""

from __future__ import annotations
from datetime import datetime
from typing import Any
from uuid import UUID
from pydantic import BaseModel, ConfigDict, Field


class ChatKitSessionRequest(BaseModel):
    """Request payload for retrieving a ChatKit client secret."""

    current_client_secret: str | None = Field(default=None, alias="currentClientSecret")
    workflow_id: UUID | None = Field(default=None, alias="workflowId")
    workflow_label: str | None = Field(default=None, alias="workflowLabel")
    metadata: dict[str, Any] = Field(default_factory=dict)
    user: dict[str, Any] | None = None
    assistant: dict[str, Any] | None = None

    model_config = ConfigDict(populate_by_name=True)


class ChatKitSessionResponse(BaseModel):
    """Response payload describing a ChatKit client session."""

    client_secret: str = Field(alias="client_secret")
    expires_at: datetime | None = None

    model_config = ConfigDict(populate_by_name=True)


class ChatKitWorkflowTriggerRequest(BaseModel):
    """Payload for triggering a workflow run from ChatKit."""

    message: str
    actor: str = Field(default="chatkit")
    client_thread_id: str | None = Field(default=None, alias="client_thread_id")
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(populate_by_name=True)



"""Core behavior tests for the PostgreSQL-backed workflow repository.

These tests use in-memory fakes to verify repository behavior without requiring
a real PostgreSQL database connection.
"""

from __future__ import annotations
import asyncio
import json
from datetime import UTC, datetime
from typing import Any
from uuid import UUID, uuid4
import pytest
from orcheo.models.workflow import WorkflowRunStatus
from orcheo.triggers.cron import CronTriggerConfig
from orcheo.triggers.retry import RetryDecision, RetryPolicyConfig
from orcheo.triggers.webhook import WebhookTriggerConfig
from orcheo_backend.app.repository.errors import (
    WorkflowNotFoundError,
    WorkflowPublishStateError,
)
from orcheo_backend.app.repository_postgres import PostgresWorkflowRepository
from orcheo_backend.app.repository_postgres import _base as pg_base


class FakeRow(dict[str, Any]):
    """Fake row that supports both index and key access like psycopg rows."""

    def __getitem__(self, key: str | int) -> Any:
        if isinstance(key, int):
            return list(self.values())[key]
        return super().__getitem__(key)


class FakeCursor:
    """Fake database cursor for testing."""

    def __init__(
        self, *, row: dict[str, Any] | None = None, rows: list[Any] | None = None
    ) -> None:
        self._row = FakeRow(row) if row else None
        self._rows = [FakeRow(r) if isinstance(r, dict) else r for r in (rows or [])]

    async def fetchone(self) -> FakeRow | None:
        return self._row

    async def fetchall(self) -> list[Any]:
        return list(self._rows)


class FakeConnection:
    """Fake database connection recording queries and returning configured responses."""

    def __init__(self, responses: list[Any]) -> None:
        self._responses = list(responses)
        self.queries: list[tuple[str, Any | None]] = []
        self.commits = 0
        self.rollbacks = 0

    async def execute(self, query: str, params: Any | None = None) -> FakeCursor:
        self.queries.append((query.strip(), params))
        response = self._responses.pop(0) if self._responses else {}
        if isinstance(response, FakeCursor):
            return response
        if isinstance(response, dict):
            return FakeCursor(row=response.get("row"), rows=response.get("rows"))
        if isinstance(response, list):
            return FakeCursor(rows=response)
        return FakeCursor()

    async def commit(self) -> None:
        self.commits += 1

    async def rollback(self) -> None:
        self.rollbacks += 1

    async def __aenter__(self) -> FakeConnection:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: Any,
    ) -> None:
        return None


class FakePool:
    """Fake connection pool that returns a pre-configured connection."""

    def __init__(self, connection: FakeConnection) -> None:
        self._connection = connection

    async def open(self) -> None:
        return None

    async def close(self) -> None:
        return None

    def connection(self) -> FakeConnection:
        return self._connection


def make_repository(
    monkeypatch: pytest.MonkeyPatch,
    responses: list[Any],
    *,
    initialized: bool = True,
) -> PostgresWorkflowRepository:
    """Create a PostgresWorkflowRepository with fake connection pool."""
    monkeypatch.setattr(pg_base, "AsyncConnectionPool", object())
    monkeypatch.setattr(pg_base, "DictRowFactory", object())
    repo = PostgresWorkflowRepository("postgresql://test")
    repo._pool = FakePool(FakeConnection(responses))
    repo._initialized = initialized
    return repo


def _workflow_payload(workflow_id: UUID, **overrides: Any) -> dict[str, Any]:
    """Generate a fake workflow payload dictionary."""
    now = datetime.now(tz=UTC).isoformat()
    base = {
        "id": str(workflow_id),
        "name": "Test Workflow",
        "slug": "test-workflow",
        "description": None,
        "tags": [],
        "is_archived": False,
        "is_public": False,
        "published_at": None,
        "published_by": None,
        "require_login": False,
        "audit_log": [],
        "created_at": now,
        "updated_at": now,
    }
    base.update(overrides)
    return base


def _version_payload(
    version_id: UUID, workflow_id: UUID, version: int = 1, **overrides: Any
) -> dict[str, Any]:
    """Generate a fake version payload dictionary."""
    now = datetime.now(tz=UTC).isoformat()
    base = {
        "id": str(version_id),
        "workflow_id": str(workflow_id),
        "version": version,
        "graph": {},
        "metadata": {},
        "runnable_config": None,
        "notes": None,
        "created_by": "author",
        "audit_log": [],
        "created_at": now,
        "updated_at": now,
    }
    base.update(overrides)
    return base


def _run_payload(
    run_id: UUID,
    version_id: UUID,
    *,
    status: str = "pending",
    triggered_by: str = "manual",
    **overrides: Any,
) -> dict[str, Any]:
    """Generate a fake run payload dictionary."""
    now = datetime.now(tz=UTC).isoformat()
    base = {
        "id": str(run_id),
        "workflow_version_id": str(version_id),
        "status": status,
        "triggered_by": triggered_by,
        "input_payload": {},
        "runnable_config": {},
        "tags": [],
        "callbacks": [],
        "metadata": {},
        "run_name": None,
        "output_payload": None,
        "error": None,
        "audit_log": [],
        "created_at": now,
        "updated_at": now,
        "started_at": None,
        "completed_at": None,
    }
    base.update(overrides)
    return base


@pytest.mark.asyncio
async def test_postgres_repository_list_workflows_excludes_archived(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing workflows excludes archived by default."""
    workflow_id_active = uuid4()
    workflow_id_archived = uuid4()

    responses: list[Any] = [
        {
            "rows": [
                {"payload": _workflow_payload(workflow_id_active)},
                {
                    "payload": _workflow_payload(
                        workflow_id_archived, is_archived=True, name="Archived"
                    )
                },
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    workflows = await repo.list_workflows()
    assert len(workflows) == 1
    assert workflows[0].id == workflow_id_active


@pytest.mark.asyncio
async def test_postgres_repository_list_workflows_includes_archived(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing workflows with include_archived=True returns all workflows."""
    workflow_id_active = uuid4()
    workflow_id_archived = uuid4()

    responses: list[Any] = [
        {
            "rows": [
                {"payload": _workflow_payload(workflow_id_active)},
                {
                    "payload": _workflow_payload(
                        workflow_id_archived, is_archived=True, name="Archived"
                    )
                },
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    workflows = await repo.list_workflows(include_archived=True)
    assert len(workflows) == 2


@pytest.mark.asyncio
async def test_postgres_repository_create_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Creating a workflow persists it to the database."""
    responses: list[Any] = [{}]
    repo = make_repository(monkeypatch, responses)

    workflow = await repo.create_workflow(
        name="Test",
        slug="test-slug",
        description="Test description",
        tags=["tag1", "tag2"],
        actor="tester",
    )

    assert workflow.name == "Test"
    assert workflow.slug == "test-slug"
    assert workflow.description == "Test description"
    assert workflow.tags == ["tag1", "tag2"]
    assert workflow.audit_log[0].action == "workflow_created"


@pytest.mark.asyncio
async def test_postgres_repository_get_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Getting a workflow by ID returns the workflow."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, name="Found")}},
    ]
    repo = make_repository(monkeypatch, responses)

    workflow = await repo.get_workflow(workflow_id)
    assert workflow.id == workflow_id
    assert workflow.name == "Found"


@pytest.mark.asyncio
async def test_postgres_repository_update_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Updating a workflow records changes and persists them."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, name="Original")}},
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    workflow = await repo.update_workflow(
        workflow_id,
        name="Updated",
        description="New description",
        tags=["new-tag"],
        is_archived=None,
        actor="updater",
    )

    assert workflow.name == "Updated"
    assert workflow.description == "New description"
    assert workflow.tags == ["new-tag"]
    assert any(e.action == "workflow_updated" for e in workflow.audit_log)


@pytest.mark.asyncio
async def test_postgres_repository_archive_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Archiving a workflow sets is_archived to True."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    workflow = await repo.archive_workflow(workflow_id, actor="archiver")
    assert workflow.is_archived is True


@pytest.mark.asyncio
async def test_postgres_repository_archive_revokes_publish(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Archiving a public workflow revokes publish state."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {
            "row": {
                "payload": _workflow_payload(
                    workflow_id,
                    is_public=True,
                    require_login=True,
                    published_by="publisher",
                )
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    workflow = await repo.archive_workflow(workflow_id, actor="archiver")

    assert workflow.is_archived is True
    assert workflow.is_public is False
    assert workflow.require_login is False


@pytest.mark.asyncio
async def test_postgres_repository_publish_revoke_lifecycle(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Publishing and revoking workflows updates public state."""
    workflow_id = uuid4()

    # First call: get workflow for publish
    # Second call: update after publish
    # Third call: get workflow for revoke
    # Fourth call: update after revoke
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},
        {
            "row": {
                "payload": _workflow_payload(
                    workflow_id, is_public=True, require_login=True
                )
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    published = await repo.publish_workflow(
        workflow_id, require_login=True, actor="publisher"
    )
    assert published.is_public is True
    assert published.require_login is True

    revoked = await repo.revoke_publish(workflow_id, actor="revoker")
    assert revoked.is_public is False


@pytest.mark.asyncio
async def test_postgres_repository_publish_archived_raises_not_found(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Publishing an archived workflow raises WorkflowNotFoundError."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, is_archived=True)}},
    ]
    repo = make_repository(monkeypatch, responses)

    with pytest.raises(WorkflowNotFoundError):
        await repo.publish_workflow(workflow_id, require_login=False, actor="actor")


@pytest.mark.asyncio
async def test_postgres_repository_revoke_archived_raises_not_found(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Revoking publish on an archived workflow raises WorkflowNotFoundError."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, is_archived=True)}},
    ]
    repo = make_repository(monkeypatch, responses)

    with pytest.raises(WorkflowNotFoundError):
        await repo.revoke_publish(workflow_id, actor="actor")


@pytest.mark.asyncio
async def test_postgres_repository_publish_already_public_raises(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Publishing an already public workflow raises WorkflowPublishStateError."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, is_public=True)}},
    ]
    repo = make_repository(monkeypatch, responses)

    with pytest.raises(WorkflowPublishStateError):
        await repo.publish_workflow(workflow_id, require_login=False, actor="publisher")


@pytest.mark.asyncio
async def test_postgres_repository_revoke_unpublished_raises(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Revoking a non-public workflow raises WorkflowPublishStateError."""
    workflow_id = uuid4()
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id, is_public=False)}},
    ]
    repo = make_repository(monkeypatch, responses)

    with pytest.raises(WorkflowPublishStateError):
        await repo.revoke_publish(workflow_id, actor="revoker")


@pytest.mark.asyncio
async def test_postgres_repository_webhook_trigger_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Configuring and retrieving webhook triggers works correctly."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT webhook_triggers
    ]
    repo = make_repository(monkeypatch, responses)

    config = await repo.configure_webhook_trigger(
        workflow_id, WebhookTriggerConfig(allowed_methods={"POST", "PUT"})
    )

    assert "POST" in config.allowed_methods
    assert "PUT" in config.allowed_methods


@pytest.mark.asyncio
async def test_postgres_repository_get_webhook_trigger_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Getting webhook trigger config returns the configured webhook."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT webhook_triggers
        {"row": {"payload": _workflow_payload(workflow_id)}},
    ]
    repo = make_repository(monkeypatch, responses)

    # Configure webhook first
    await repo.configure_webhook_trigger(
        workflow_id, WebhookTriggerConfig(allowed_methods={"POST"})
    )

    # Get the config
    config = await repo.get_webhook_trigger_config(workflow_id)
    assert "POST" in config.allowed_methods


@pytest.mark.asyncio
async def test_postgres_repository_cron_trigger_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Configuring and retrieving cron triggers works correctly."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT cron_triggers
    ]
    repo = make_repository(monkeypatch, responses)

    config = await repo.configure_cron_trigger(
        workflow_id, CronTriggerConfig(expression="0 9 * * *", timezone="UTC")
    )

    assert config.expression == "0 9 * * *"
    assert config.timezone == "UTC"


@pytest.mark.asyncio
async def test_postgres_repository_get_cron_trigger_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Getting cron trigger config returns the configured cron."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT cron_triggers
        {"row": {"payload": _workflow_payload(workflow_id)}},
    ]
    repo = make_repository(monkeypatch, responses)

    # Configure cron first
    await repo.configure_cron_trigger(
        workflow_id, CronTriggerConfig(expression="0 9 * * *", timezone="UTC")
    )

    # Get the config
    config = await repo.get_cron_trigger_config(workflow_id)
    assert config.expression == "0 9 * * *"
    assert config.timezone == "UTC"


@pytest.mark.asyncio
async def test_postgres_repository_retry_policy_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Configuring and retrieving retry policies works correctly."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT retry_policies
    ]
    repo = make_repository(monkeypatch, responses)

    config = await repo.configure_retry_policy(
        workflow_id,
        RetryPolicyConfig(max_attempts=3, initial_delay_seconds=1.0, jitter_factor=0.1),
    )

    assert config.max_attempts == 3


@pytest.mark.asyncio
async def test_postgres_repository_create_version(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Creating a version assigns an auto-incremented version number."""
    workflow_id = uuid4()

    responses: list[Any] = [
        # get_workflow for create_version
        {"row": {"payload": _workflow_payload(workflow_id)}},
        # SELECT max version
        {"row": {"max_version": None}},
        # INSERT workflow_versions
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    version = await repo.create_version(
        workflow_id,
        graph={"nodes": []},
        metadata={"key": "value"},
        notes="Initial version",
        created_by="author",
    )

    assert version.workflow_id == workflow_id
    assert version.version == 1
    assert version.graph == {"nodes": []}
    assert version.metadata == {"key": "value"}


@pytest.mark.asyncio
async def test_postgres_repository_list_versions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing versions returns versions ordered by version number."""
    workflow_id = uuid4()
    version_id_1 = uuid4()
    version_id_2 = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "rows": [
                {"payload": _version_payload(version_id_1, workflow_id, version=1)},
                {"payload": _version_payload(version_id_2, workflow_id, version=2)},
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    versions = await repo.list_versions(workflow_id)

    assert len(versions) == 2
    assert versions[0].version == 1
    assert versions[1].version == 2


@pytest.mark.asyncio
async def test_postgres_repository_list_versions_with_json_string_payload(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing versions handles JSON string payloads correctly."""
    workflow_id = uuid4()
    version_id = uuid4()

    # Some PostgreSQL drivers return JSON as strings
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "rows": [
                {"payload": json.dumps(_version_payload(version_id, workflow_id))},
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    versions = await repo.list_versions(workflow_id)

    assert len(versions) == 1
    assert versions[0].id == version_id


@pytest.mark.asyncio
async def test_postgres_repository_get_latest_version_with_json_string(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Getting latest version handles JSON string payloads correctly."""
    workflow_id = uuid4()
    version_id = uuid4()

    # Some PostgreSQL drivers return JSON as strings
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {"row": {"payload": json.dumps(_version_payload(version_id, workflow_id))}},
    ]
    repo = make_repository(monkeypatch, responses)

    version = await repo.get_latest_version(workflow_id)

    assert version.id == version_id


@pytest.mark.asyncio
async def test_postgres_repository_list_runs_for_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing runs for a workflow returns runs ordered by creation time."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id_1 = uuid4()
    run_id_2 = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "rows": [
                {"payload": _run_payload(run_id_1, version_id)},
                {"payload": _run_payload(run_id_2, version_id)},
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    runs = await repo.list_runs_for_workflow(workflow_id)

    assert len(runs) == 2
    assert runs[0].id == run_id_1
    assert runs[1].id == run_id_2


@pytest.mark.asyncio
async def test_postgres_repository_list_runs_with_json_string_payload(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Listing runs handles JSON string payloads correctly."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    # Some PostgreSQL drivers return JSON as strings
    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "rows": [
                {"payload": json.dumps(_run_payload(run_id, version_id))},
            ]
        },
    ]
    repo = make_repository(monkeypatch, responses)

    runs = await repo.list_runs_for_workflow(workflow_id)

    assert len(runs) == 1
    assert runs[0].id == run_id


@pytest.mark.asyncio
async def test_postgres_repository_run_lifecycle_transitions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Run status transitions work correctly (started -> succeeded/failed)."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    # For mark_run_started: get run, update
    responses: list[Any] = [
        {
            "row": {
                "payload": _run_payload(run_id, version_id, status="pending"),
                "workflow_id": str(workflow_id),
                "triggered_by": "manual",
                "status": "pending",
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    run = await repo.mark_run_started(run_id, actor="worker")
    assert run.status == WorkflowRunStatus.RUNNING


@pytest.mark.asyncio
async def test_postgres_repository_mark_run_succeeded(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Marking a run as succeeded sets status and output."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    responses: list[Any] = [
        {
            "row": {
                "payload": _run_payload(run_id, version_id, status="running"),
                "workflow_id": str(workflow_id),
                "triggered_by": "manual",
                "status": "running",
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    run = await repo.mark_run_succeeded(
        run_id, actor="worker", output={"result": "done"}
    )
    assert run.status == WorkflowRunStatus.SUCCEEDED
    assert run.output_payload == {"result": "done"}


@pytest.mark.asyncio
async def test_postgres_repository_mark_run_failed(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Marking a run as failed sets status and error."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    responses: list[Any] = [
        {
            "row": {
                "payload": _run_payload(run_id, version_id, status="running"),
                "workflow_id": str(workflow_id),
                "triggered_by": "manual",
                "status": "running",
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    run = await repo.mark_run_failed(run_id, actor="worker", error="Something broke")
    assert run.status == WorkflowRunStatus.FAILED
    assert run.error == "Something broke"


@pytest.mark.asyncio
async def test_postgres_repository_mark_run_cancelled(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Marking a run as cancelled sets status and reason."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    responses: list[Any] = [
        {
            "row": {
                "payload": _run_payload(run_id, version_id, status="pending"),
                "workflow_id": str(workflow_id),
                "triggered_by": "manual",
                "status": "pending",
            }
        },
        {},
    ]
    repo = make_repository(monkeypatch, responses)

    run = await repo.mark_run_cancelled(run_id, actor="user", reason="No longer needed")
    assert run.status == WorkflowRunStatus.CANCELLED


@pytest.mark.asyncio
async def test_postgres_repository_reset_clears_all_data(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Reset clears all workflow data from the repository."""
    # Reset executes 6 DELETE statements
    responses: list[Any] = [{}, {}, {}, {}, {}, {}]
    repo = make_repository(monkeypatch, responses)

    await repo.reset()

    # Verify trigger layer was reset
    assert len(repo._trigger_layer._cron_states) == 0
    assert len(repo._trigger_layer._webhook_states) == 0
    assert len(repo._trigger_layer._retry_configs) == 0


@pytest.mark.asyncio
async def test_postgres_repository_get_retry_policy_config(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Getting retry policy config returns the configured policy."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT retry_policies
        {"row": {"payload": _workflow_payload(workflow_id)}},
    ]
    repo = make_repository(monkeypatch, responses)

    # Configure retry policy first
    await repo.configure_retry_policy(
        workflow_id,
        RetryPolicyConfig(max_attempts=3, initial_delay_seconds=1.0, jitter_factor=0.1),
    )

    # Get the policy
    config = await repo.get_retry_policy_config(workflow_id)
    assert config.max_attempts == 3
    assert config.initial_delay_seconds == 1.0
    assert config.jitter_factor == 0.1


@pytest.mark.asyncio
async def test_postgres_repository_schedule_retry_for_run(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Scheduling a retry for a failed run returns a retry decision."""
    workflow_id = uuid4()
    version_id = uuid4()
    run_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT retry_policies
        {
            "row": {
                "payload": _run_payload(run_id, version_id, status="failed"),
                "workflow_id": str(workflow_id),
                "triggered_by": "manual",
                "status": "failed",
            }
        },
    ]
    repo = make_repository(monkeypatch, responses)

    # Configure retry policy first
    await repo.configure_retry_policy(
        workflow_id,
        RetryPolicyConfig(max_attempts=3, initial_delay_seconds=1.0, jitter_factor=0.0),
    )

    # Track the run so retry can find its workflow
    repo._trigger_layer.track_run(workflow_id, run_id)

    # Schedule retry
    decision = await repo.schedule_retry_for_run(run_id)
    assert decision is not None
    assert isinstance(decision, RetryDecision)
    assert decision.retry_number == 1


@pytest.mark.asyncio
async def test_postgres_repository_ensure_initialized_runs_once(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Initialization only runs once even with concurrent calls."""
    repo = make_repository(monkeypatch, [], initialized=True)

    await asyncio.gather(
        repo._ensure_initialized(),
        repo._ensure_initialized(),
    )

    assert repo._initialized is True


@pytest.mark.asyncio
async def test_postgres_repository_delete_cron_trigger(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Deleting a cron trigger removes it from the database and trigger layer."""
    workflow_id = uuid4()

    responses: list[Any] = [
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # INSERT cron_triggers
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {},  # DELETE cron_triggers
    ]
    repo = make_repository(monkeypatch, responses)

    # Configure first
    await repo.configure_cron_trigger(
        workflow_id, CronTriggerConfig(expression="0 9 * * *", timezone="UTC")
    )

    # Delete
    await repo.delete_cron_trigger(workflow_id)

    # Verify trigger layer no longer has the config
    assert workflow_id not in repo._trigger_layer._cron_states


@pytest.mark.asyncio
async def test_postgres_repository_diff_versions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Diffing versions returns a unified diff."""
    workflow_id = uuid4()
    version_id_1 = uuid4()
    version_id_2 = uuid4()

    responses: list[Any] = [
        # get_version_by_number for base_version
        # (calls get_workflow then fetches version)
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "row": {
                "payload": _version_payload(
                    version_id_1, workflow_id, version=1, graph={"nodes": ["a"]}
                )
            }
        },
        # get_version_by_number for target_version
        # (calls get_workflow then fetches version)
        {"row": {"payload": _workflow_payload(workflow_id)}},
        {
            "row": {
                "payload": _version_payload(
                    version_id_2, workflow_id, version=2, graph={"nodes": ["a", "b"]}
                )
            }
        },
    ]
    repo = make_repository(monkeypatch, responses)

    diff_result = await repo.diff_versions(
        workflow_id, base_version=1, target_version=2
    )

    # Diff result contains the version numbers and a diff list
    assert diff_result.base_version == 1
    assert diff_result.target_version == 2
    # The diff list should contain changes
    assert len(diff_result.diff) > 0


@pytest.mark.asyncio
async def test_base_repository_dependency_check(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(pg_base, "AsyncConnectionPool", None)
    monkeypatch.setattr(pg_base, "DictRowFactory", None)

    with pytest.raises(RuntimeError, match="psycopg"):
        PostgresWorkflowRepository("postgresql://test")


@pytest.mark.asyncio
async def test_base_repository_get_pool_race(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(pg_base, "AsyncConnectionPool", object())
    monkeypatch.setattr(pg_base, "DictRowFactory", object())
    repo = PostgresWorkflowRepository("postgresql://test")
    # Simulate pool created by another task
    repo._pool = "existing"  # type: ignore

    pool = await repo._get_pool()
    assert pool == "existing"

    # Simulate race
    repo._pool = None

    class SideEffectLock:
        async def __aenter__(self):
            repo._pool = "race_pool"  # type: ignore

        async def __aexit__(self, *args):
            pass

    repo._pool_lock = SideEffectLock()  # type: ignore

    pool = await repo._get_pool()
    assert pool == "race_pool"


@pytest.mark.asyncio
async def test_base_repository_ensure_initialized_race(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    repo = make_repository(monkeypatch, [], initialized=False)

    class SideEffectLock:
        async def __aenter__(self):
            repo._initialized = True

        async def __aexit__(self, *args):
            pass

    repo._init_lock = SideEffectLock()  # type: ignore

    # Should return early
    await repo._ensure_initialized()
    assert repo._initialized


@pytest.mark.asyncio
async def test_hydrate_trigger_state(monkeypatch: pytest.MonkeyPatch) -> None:
    # We provide rows for retry_policies, webhook_triggers, cron_triggers,
    # and workflow_runs
    w_id = uuid4()
    retry_conf = RetryPolicyConfig(max_attempts=5).model_dump(mode="json")
    webhook_conf = WebhookTriggerConfig(allowed_methods={"GET"}).model_dump(mode="json")
    cron_conf = CronTriggerConfig(expression="* * * * *", timezone="UTC").model_dump(
        mode="json"
    )

    # We need to skip schema exec in _ensure_initialized implicitly by mocking
    # _connection to consume schema statements
    # Actually make_repository sets initialized=True. We want False here so
    # _ensure_initialized runs.

    # Let's manually construct simpler responses since exact schema statement
    # count varies
    # We can mock _connection to just yield a conn that handles the queries
    # we care about

    repo = make_repository(monkeypatch, [], initialized=False)

    # Mock connection to ignore schema statements (execute returns nothing)
    # Then for hydrate queries return specific data

    class HydrateMockValues:
        def pop_response(self, query):
            if "information_schema.columns" in query:
                return []
            if "retry_policies" in query:
                return [{"workflow_id": str(w_id), "config": retry_conf}]
            if "webhook_triggers" in query:
                return [{"workflow_id": str(w_id), "config": webhook_conf}]
            if "FROM cron_triggers" in query:
                return [
                    {
                        "workflow_id": str(w_id),
                        "config": cron_conf,
                        "last_dispatched_at": None,
                    }
                ]
            if "workflow_runs" in query:
                return []
            return []

    hydrator = HydrateMockValues()

    class SmartFakeConnection(FakeConnection):
        async def execute(self, query, params=None):
            self.queries.append((query, params))
            rows = hydrator.pop_response(query)
            return FakeCursor(rows=rows)

    repo._pool = FakePool(SmartFakeConnection([]))  # type: ignore
    repo._initialized = False  # Force init

    await repo._ensure_initialized()

    # Check trigger layer state
    assert w_id in repo._trigger_layer._retry_configs
    assert repo._trigger_layer._retry_configs[w_id].max_attempts == 5
    assert w_id in repo._trigger_layer._webhook_states
    assert w_id in repo._trigger_layer._cron_states


@pytest.mark.asyncio
async def test_refresh_cron_triggers_sync(monkeypatch: pytest.MonkeyPatch) -> None:
    repo = make_repository(monkeypatch, [], initialized=True)
    w_id = uuid4()
    # Initial state: no cron
    assert w_id not in repo._trigger_layer._cron_states

    # 1. Add cron
    cron_conf = CronTriggerConfig(expression="* * * * *", timezone="UTC")
    responses = [
        {
            "rows": [
                {
                    "workflow_id": str(w_id),
                    "config": cron_conf.model_dump(mode="json"),
                    "last_dispatched_at": None,
                }
            ]
        }
    ]
    repo._pool = FakePool(FakeConnection(responses))  # type: ignore

    await repo._refresh_cron_triggers()
    assert w_id in repo._trigger_layer._cron_states

    # 2. Update cron
    cron_conf_2 = CronTriggerConfig(expression="0 0 * * *", timezone="UTC")
    responses = [
        {
            "rows": [
                {
                    "workflow_id": str(w_id),
                    "config": cron_conf_2.model_dump(mode="json"),
                    "last_dispatched_at": None,
                }
            ]
        }
    ]
    repo._pool = FakePool(FakeConnection(responses))  # type: ignore

    await repo._refresh_cron_triggers()
    stored = repo._trigger_layer._cron_states[w_id].config
    assert stored.expression == "0 0 * * *"

    # 2b. Update last_dispatched_at without changing config
    last_dispatched_at = datetime(2025, 1, 1, 9, 0, tzinfo=UTC)
    responses = [
        {
            "rows": [
                {
                    "workflow_id": str(w_id),
                    "config": cron_conf_2.model_dump(mode="json"),
                    "last_dispatched_at": last_dispatched_at,
                }
            ]
        }
    ]
    repo._pool = FakePool(FakeConnection(responses))  # type: ignore

    await repo._refresh_cron_triggers()
    stored_state = repo._trigger_layer._cron_states[w_id]
    assert stored_state.last_dispatched_at == last_dispatched_at

    # 3. Remove cron
    responses = [{"rows": []}]
    repo._pool = FakePool(FakeConnection(responses))  # type: ignore

    await repo._refresh_cron_triggers()
    assert w_id not in repo._trigger_layer._cron_states



"""Coverage for the public workflow router endpoint."""

from __future__ import annotations
from uuid import UUID, uuid4
import pytest
from fastapi import HTTPException
from orcheo.models.workflow import Workflow
from orcheo_backend.app.repository import WorkflowNotFoundError
from orcheo_backend.app.routers import workflows
from orcheo_backend.app.routers.workflows import _resolve_chatkit_public_base_url


class _WorkflowRepo:
    def __init__(self, workflow: Workflow) -> None:
        self.workflow = workflow

    async def get_workflow(self, workflow_id: UUID) -> Workflow:
        if workflow_id != self.workflow.id:
            raise WorkflowNotFoundError(str(workflow_id))
        return self.workflow


class _MissingWorkflowRepo:
    async def get_workflow(self, workflow_id: UUID) -> Workflow:
        raise WorkflowNotFoundError(str(workflow_id))


@pytest.mark.asyncio()
async def test_get_public_workflow_not_found() -> None:
    with pytest.raises(HTTPException) as excinfo:
        await workflows.get_public_workflow(
            uuid4(),
            _MissingWorkflowRepo(),
        )

    assert excinfo.value.status_code == 404


@pytest.mark.asyncio()
async def test_get_public_workflow_denies_private_workflows() -> None:
    workflow = Workflow(name="Hidden workflow", is_public=False)
    repo = _WorkflowRepo(workflow)

    with pytest.raises(HTTPException) as excinfo:
        await workflows.get_public_workflow(workflow.id, repo)

    assert excinfo.value.status_code == 403
    assert excinfo.value.detail["code"] == "workflow.not_public"


@pytest.mark.asyncio()
async def test_get_public_workflow_includes_share_url(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workflow = Workflow(name="Published workflow", is_public=True)
    repo = _WorkflowRepo(workflow)
    monkeypatch.setattr(
        workflows,
        "_resolve_chatkit_public_base_url",
        lambda: "https://canvas.example",
    )

    response = await workflows.get_public_workflow(workflow.id, repo)

    assert response.share_url == f"https://canvas.example/chat/{workflow.id}"


def test_resolve_chatkit_public_base_url_returns_none_when_empty(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Verify _resolve_chatkit_public_base_url returns None when setting is empty."""
    monkeypatch.setattr(
        workflows,
        "get_settings",
        lambda: {"CHATKIT_PUBLIC_BASE_URL": ""},
    )

    result = _resolve_chatkit_public_base_url()
    assert result is None



"""Tests for ChatKit background maintenance and store helpers."""

from __future__ import annotations
import asyncio
import sqlite3
from contextlib import suppress
from datetime import UTC, datetime, timedelta
from pathlib import Path
from unittest.mock import MagicMock, patch
import pytest
from orcheo_backend.app import (
    _cancel_chatkit_cleanup_task,
    _chatkit_cleanup_task,
    _chatkit_server_ref,
    _ensure_chatkit_cleanup_task,
    _get_chatkit_store,
)
from orcheo_backend.app.chatkit import InMemoryChatKitStore
from orcheo_backend.app.chatkit_store_sqlite import SqliteChatKitStore


@pytest.mark.asyncio()
async def test_ensure_chatkit_cleanup_task_already_running() -> None:
    """Cleanup task should not be recreated if already running."""

    task = asyncio.create_task(asyncio.sleep(10))
    _chatkit_cleanup_task["task"] = task

    try:
        await _ensure_chatkit_cleanup_task()
        assert _chatkit_cleanup_task["task"] is task
    finally:
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task
        _chatkit_cleanup_task["task"] = None


@pytest.mark.asyncio()
async def test_cancel_chatkit_cleanup_task_no_task() -> None:
    """Canceling cleanup task when none exists should be safe."""

    _chatkit_cleanup_task["task"] = None
    await _cancel_chatkit_cleanup_task()
    assert _chatkit_cleanup_task["task"] is None


@pytest.mark.asyncio()
async def test_chatkit_cleanup_task_prunes_threads(tmp_path: Path) -> None:
    """Cleanup task should prune old threads and log the count."""

    db_path = tmp_path / "chatkit_test.sqlite"
    store = SqliteChatKitStore(db_path)

    mock_server = MagicMock()
    mock_server.store = store
    _chatkit_server_ref["server"] = mock_server

    from chatkit.types import ThreadMetadata

    thread_id = "thr_old"
    old_thread = ThreadMetadata(
        id=thread_id,
        created_at=datetime.now(tz=UTC) - timedelta(days=60),
    )
    await store.save_thread(old_thread, {})

    with sqlite3.connect(db_path) as conn:
        stale_timestamp = (datetime.now(tz=UTC) - timedelta(days=60)).isoformat()
        conn.execute(
            "UPDATE chat_threads SET updated_at = ? WHERE id = ?",
            (stale_timestamp, thread_id),
        )
        conn.commit()

    with (
        patch("orcheo_backend.app._chatkit_retention_days", return_value=30),
        patch("orcheo_backend.app._CHATKIT_CLEANUP_INTERVAL_SECONDS", 0.1),
    ):
        _chatkit_cleanup_task["task"] = None
        await _ensure_chatkit_cleanup_task()
        task = _chatkit_cleanup_task["task"]
        assert task is not None

        await asyncio.sleep(0.2)

        task.cancel()
        with suppress(asyncio.CancelledError):
            await task

    _chatkit_server_ref["server"] = None
    _chatkit_cleanup_task["task"] = None


def test_get_chatkit_store_returns_none_when_no_server() -> None:
    """Get chatkit store should return None when server is not initialized."""

    _chatkit_server_ref["server"] = None
    store = _get_chatkit_store()
    assert store is None


def test_get_chatkit_store_returns_none_for_non_sqlite_store() -> None:
    """Get chatkit store should return None when store is not SqliteChatKitStore."""

    mock_server = MagicMock()
    mock_server.store = InMemoryChatKitStore()
    _chatkit_server_ref["server"] = mock_server

    try:
        store = _get_chatkit_store()
        assert store is None
    finally:
        _chatkit_server_ref["server"] = None



"""Metadata handling and attachment inference tests for the SQLite ChatKit store."""

from __future__ import annotations
from datetime import UTC, datetime
from pathlib import Path
from types import SimpleNamespace
import pytest
from chatkit.types import FileAttachment, ThreadMetadata
from orcheo_backend.app.chatkit_store_sqlite import SqliteChatKitStore


def _timestamp() -> datetime:
    return datetime.now(tz=UTC)


@pytest.mark.asyncio
async def test_sqlite_store_merges_metadata_from_context(tmp_path: Path) -> None:
    """Incoming ChatKit metadata should populate the stored thread."""

    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    request = SimpleNamespace(
        metadata={"workflow_id": "wf_ctx", "workflow_name": "Ctx"}
    )
    context: dict[str, object] = {"chatkit_request": request}

    thread = ThreadMetadata(
        id="thr_ctx",
        created_at=_timestamp(),
    )

    await store.save_thread(thread, context)

    assert thread.metadata["workflow_id"] == "wf_ctx"

    loaded_thread = await store.load_thread(thread.id, {})
    assert loaded_thread.metadata["workflow_id"] == "wf_ctx"


@pytest.mark.asyncio
async def test_sqlite_store_merge_metadata_no_request(tmp_path: Path) -> None:
    """Save thread should handle context without chatkit_request."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    thread = ThreadMetadata(
        id="thr_no_req",
        created_at=_timestamp(),
        metadata={"existing": "value"},
    )

    context: dict[str, object] = {"other_key": "other_value"}
    await store.save_thread(thread, context)

    loaded = await store.load_thread(thread.id, {})
    assert loaded.metadata["existing"] == "value"


@pytest.mark.asyncio
async def test_sqlite_store_merge_metadata_no_metadata_attr(tmp_path: Path) -> None:
    """Save thread should handle request without metadata attribute."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    request = SimpleNamespace(params="test")
    context: dict[str, object] = {"chatkit_request": request}

    thread = ThreadMetadata(
        id="thr_no_meta",
        created_at=_timestamp(),
        metadata={"existing": "value"},
    )

    await store.save_thread(thread, context)

    loaded = await store.load_thread(thread.id, {})
    assert loaded.metadata["existing"] == "value"


@pytest.mark.asyncio
async def test_sqlite_store_merge_metadata_empty_dict(tmp_path: Path) -> None:
    """Save thread should handle request with empty metadata dict."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    request = SimpleNamespace(metadata={})
    context: dict[str, object] = {"chatkit_request": request}

    thread = ThreadMetadata(
        id="thr_empty_meta",
        created_at=_timestamp(),
        metadata={"existing": "value"},
    )

    await store.save_thread(thread, context)

    loaded = await store.load_thread(thread.id, {})
    assert loaded.metadata["existing"] == "value"


@pytest.mark.asyncio
async def test_sqlite_store_merge_metadata_non_dict_metadata(tmp_path: Path) -> None:
    """Save thread should handle request with non-dict metadata."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    request = SimpleNamespace(metadata="not a dict")
    context: dict[str, object] = {"chatkit_request": request}

    thread = ThreadMetadata(
        id="thr_non_dict_meta",
        created_at=_timestamp(),
        metadata={"existing": "value"},
    )

    await store.save_thread(thread, context)

    loaded = await store.load_thread(thread.id, {})
    assert loaded.metadata["existing"] == "value"


@pytest.mark.asyncio
async def test_sqlite_store_infer_thread_id_from_context(tmp_path: Path) -> None:
    """Save attachment should infer thread_id from context."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    thread = ThreadMetadata(id="thr_infer", created_at=_timestamp())
    await store.save_thread(thread, {})

    params = SimpleNamespace(thread_id="thr_infer")
    request = SimpleNamespace(params=params)
    context: dict[str, object] = {"chatkit_request": request}

    attachment = FileAttachment(
        id="atc_infer",
        name="test.txt",
        mime_type="text/plain",
    )
    await store.save_attachment(attachment, context)

    loaded = await store.load_attachment(attachment.id, {})
    assert loaded.name == "test.txt"


@pytest.mark.asyncio
async def test_sqlite_store_infer_thread_id_no_context(tmp_path: Path) -> None:
    """Infer thread ID should return None when context is missing."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    attachment = FileAttachment(
        id="atc_no_ctx",
        name="test.txt",
        mime_type="text/plain",
    )

    context: dict[str, object] = {}
    await store.save_attachment(attachment, context)

    loaded = await store.load_attachment(attachment.id, context)
    assert loaded.name == "test.txt"


@pytest.mark.asyncio
async def test_sqlite_store_infer_thread_id_no_params(tmp_path: Path) -> None:
    """Infer thread ID should handle request without params."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    request = SimpleNamespace()
    context: dict[str, object] = {"chatkit_request": request}

    attachment = FileAttachment(
        id="atc_no_params",
        name="test.txt",
        mime_type="text/plain",
    )

    await store.save_attachment(attachment, context)

    loaded = await store.load_attachment(attachment.id, context)
    assert loaded.name == "test.txt"


@pytest.mark.asyncio
async def test_sqlite_store_persists_storage_path(tmp_path: Path) -> None:
    """Save attachment should persist storage_path for cleanup and retrieval."""
    db_path = tmp_path / "store.sqlite"
    store = SqliteChatKitStore(db_path)

    attachment = FileAttachment(
        id="atc_path",
        name="file.txt",
        mime_type="text/plain",
    )
    storage_path = tmp_path / "uploads" / "atc_path_file.txt"

    await store.save_attachment(
        attachment,
        context={},
        storage_path=str(storage_path),
    )

    async with store._connection() as conn:  # type: ignore[attr-defined]  # noqa: SLF001
        cursor = await conn.execute(
            "SELECT storage_path FROM chat_attachments WHERE id = ?",
            (attachment.id,),
        )
        row = await cursor.fetchone()

    assert row is not None
    assert row["storage_path"] == str(storage_path)



"""Unit tests for conversational search ingestion models."""

import pytest
from orcheo.nodes.conversational_search.models import Document, DocumentChunk


def test_document_trims_content_and_rejects_empty_strings() -> None:
    document = Document(id="doc", content="  trimmed  ")
    assert document.content == "trimmed"

    with pytest.raises(
        ValueError, match="Document content cannot be empty after trimming whitespace"
    ):
        Document(id="doc", content="   ")


def test_document_chunk_token_count_and_validation() -> None:
    chunk = DocumentChunk(
        id="chunk-1",
        document_id="doc-1",
        index=0,
        content="  one two  ",
    )
    assert chunk.token_count == 2

    with pytest.raises(
        ValueError, match="Chunk content cannot be empty after trimming whitespace"
    ):
        DocumentChunk(id="chunk-empty", document_id="doc-1", index=0, content="   ")



"""Manual dispatch helpers for the trigger layer."""

from __future__ import annotations
from uuid import UUID
from orcheo.triggers.layer.models import ManualDispatchPlan
from orcheo.triggers.layer.state import TriggerLayerState
from orcheo.triggers.manual import ManualDispatchRequest


class ManualDispatchMixin(TriggerLayerState):
    """Provide manual dispatch planning helpers."""

    def prepare_manual_dispatch(
        self, request: ManualDispatchRequest, *, default_workflow_version_id: UUID
    ) -> ManualDispatchPlan:
        """Resolve manual dispatch runs and return the dispatch plan."""
        if request is None:
            raise ValueError("request cannot be None")
        if default_workflow_version_id is None:
            raise ValueError("default_workflow_version_id cannot be None")

        try:
            resolved_runs = request.resolve_runs(
                default_workflow_version_id=default_workflow_version_id
            )
            return ManualDispatchPlan(
                triggered_by=request.trigger_label(),
                actor=request.actor,
                runs=resolved_runs,
            )
        except Exception as exc:  # pragma: no cover - logging path
            self._logger.error("Failed to prepare manual dispatch: %s", exc)
            raise


__all__ = ["ManualDispatchMixin"]



"""Shared helpers for backend API end-to-end tests."""

from __future__ import annotations
import importlib
from datetime import UTC, datetime, timedelta
from typing import Any
from fastapi.testclient import TestClient
from orcheo.models import CredentialHealthStatus
from orcheo.vault.oauth import (
    OAuthProvider,
    OAuthTokenSecrets,
    OAuthValidationResult,
)


backend_app = importlib.import_module("orcheo_backend.app")


class StaticProvider(OAuthProvider):
    """Simple OAuth provider that returns predetermined validation results."""

    def __init__(
        self,
        *,
        status: CredentialHealthStatus = CredentialHealthStatus.HEALTHY,
        failure_reason: str | None = None,
    ) -> None:
        self.status = status
        self.failure_reason = failure_reason
        self.refresh_calls = 0

    async def refresh_tokens(
        self, metadata: Any, tokens: OAuthTokenSecrets
    ) -> OAuthTokenSecrets:
        """Return refreshed tokens recorded for testing."""

        self.refresh_calls += 1
        return OAuthTokenSecrets(
            access_token="refreshed-token",
            refresh_token="refresh-token",
            expires_at=datetime.now(tz=UTC) + timedelta(hours=1),
        )

    async def validate_tokens(
        self, metadata: Any, tokens: OAuthTokenSecrets
    ) -> OAuthValidationResult:
        """Return the configured validation status."""

        return OAuthValidationResult(
            status=self.status,
            failure_reason=self.failure_reason,
        )


def create_workflow_with_version(api_client: TestClient) -> tuple[str, str]:
    """Create a workflow along with an initial version."""

    workflow_response = api_client.post(
        "/api/workflows",
        json={"name": "Webhook Flow", "actor": "tester"},
    )
    workflow_response.raise_for_status()
    workflow_id = workflow_response.json()["id"]

    version_response = api_client.post(
        f"/api/workflows/{workflow_id}/versions",
        json={
            "graph": {"nodes": ["start"], "edges": []},
            "metadata": {},
            "created_by": "tester",
        },
    )
    version_response.raise_for_status()
    version_id = version_response.json()["id"]

    return workflow_id, version_id



"""Implementation of :class:`DataTransformNode` and transform helpers."""

from __future__ import annotations
from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal
from langchain_core.runnables import RunnableConfig
from pydantic import BaseModel, Field
from orcheo.graph.state import State
from orcheo.nodes.base import TaskNode
from orcheo.nodes.data.utils import _assign_path, _extract_value
from orcheo.nodes.registry import NodeMetadata, registry


class FieldTransform(BaseModel):
    """Mapping definition used by :class:`DataTransformNode`."""

    source: str | None = Field(
        default=None, description="Dotted path pointing to the source value"
    )
    target: str = Field(description="Dotted path where the transformed value is stored")
    default: Any | None = Field(
        default=None, description="Default value used when the source is missing"
    )
    when_missing: Literal["default", "skip"] = Field(
        default="default",
        description="Control whether missing values use the default or are skipped",
    )
    transform: Literal[
        "identity",
        "string",
        "int",
        "float",
        "bool",
        "lower",
        "upper",
        "title",
        "length",
    ] = Field(
        default="identity",
        description="Optional transformation applied to the extracted value",
    )


def _transform_identity(value: Any) -> Any:
    return value


def _transform_string(value: Any) -> str:
    return "" if value is None else str(value)


def _transform_int(value: Any) -> int:
    return 0 if value is None else int(value)


def _transform_float(value: Any) -> float:
    return 0.0 if value is None else float(value)


def _transform_bool(value: Any) -> bool:
    return bool(value)


def _transform_lower(value: Any) -> Any:
    return value.lower() if isinstance(value, str) else value


def _transform_upper(value: Any) -> Any:
    return value.upper() if isinstance(value, str) else value


def _transform_title(value: Any) -> Any:
    return value.title() if isinstance(value, str) else value


def _transform_length(value: Any) -> int:
    if isinstance(value, Mapping):
        return len(value)
    if isinstance(value, str | Sequence) and not isinstance(value, bytes | bytearray):
        return len(value)
    return 0


_TRANSFORM_HANDLERS: dict[str, Callable[[Any], Any]] = {
    "identity": _transform_identity,
    "string": _transform_string,
    "int": _transform_int,
    "float": _transform_float,
    "bool": _transform_bool,
    "lower": _transform_lower,
    "upper": _transform_upper,
    "title": _transform_title,
    "length": _transform_length,
}


def _apply_transform(value: Any, transform: str) -> Any:
    """Apply the requested transformation to ``value``."""
    handler = _TRANSFORM_HANDLERS.get(transform)
    if handler is None:
        return value
    return handler(value)


@registry.register(
    NodeMetadata(
        name="DataTransformNode",
        description="Map values from an input payload into a transformed structure.",
        category="data",
    )
)
class DataTransformNode(TaskNode):
    """Apply field mappings and simple transformations to structured data."""

    input_data: Any = Field(default_factory=dict, description="Input payload")
    transforms: list[FieldTransform] = Field(
        default_factory=list,
        description="Collection of field transforms applied sequentially",
    )

    async def run(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Return the transformed payload according to configured mappings."""
        output: dict[str, Any] = {}
        for mapping in self.transforms:
            if mapping.source is None:
                value = mapping.default
                found = mapping.when_missing != "skip"
            else:
                found, value = _extract_value(self.input_data, mapping.source)
                if not found:
                    if mapping.when_missing == "skip":
                        continue
                    value = mapping.default

            value = _apply_transform(value, mapping.transform)
            _assign_path(output, mapping.target, value)

        return {"result": output}



"""PostgreSQL-backed workflow repository split across focused mixins."""

from __future__ import annotations
from orcheo.vault.oauth import OAuthCredentialService
from orcheo_backend.app.repository_postgres._retry import RetryPolicyMixin
from orcheo_backend.app.repository_postgres._runs import WorkflowRunMixin
from orcheo_backend.app.repository_postgres._triggers import TriggerRepositoryMixin
from orcheo_backend.app.repository_postgres._versions import WorkflowVersionMixin
from orcheo_backend.app.repository_postgres._workflows import WorkflowRepositoryMixin


class PostgresWorkflowRepository(
    TriggerRepositoryMixin,
    RetryPolicyMixin,
    WorkflowRunMixin,
    WorkflowVersionMixin,
    WorkflowRepositoryMixin,
):
    """PostgreSQL-backed workflow repository for production deployments."""

    def __init__(
        self,
        dsn: str,
        *,
        credential_service: OAuthCredentialService | None = None,
        pool_min_size: int = 1,
        pool_max_size: int = 10,
        pool_timeout: float = 30.0,
        pool_max_idle: float = 300.0,
    ) -> None:
        """Configure repository dependencies and initialize trigger layer."""
        super().__init__(
            dsn,
            credential_service=credential_service,
            pool_min_size=pool_min_size,
            pool_max_size=pool_max_size,
            pool_timeout=pool_timeout,
            pool_max_idle=pool_max_idle,
        )


__all__ = ["PostgresWorkflowRepository"]



"""Service token MCP tools."""

from __future__ import annotations
from typing import Any
from orcheo_sdk.mcp_server.config import get_api_client
from orcheo_sdk.services import (
    create_service_token_data,
    list_service_tokens_data,
    revoke_service_token_data,
    rotate_service_token_data,
    show_service_token_data,
)


def list_service_tokens(
    profile: str | None = None,
) -> dict[str, Any]:
    """List all service tokens."""
    client, _ = get_api_client(profile=profile)
    return list_service_tokens_data(client)


def show_service_token(
    token_id: str,
    profile: str | None = None,
) -> dict[str, Any]:
    """Display details for a specific service token."""
    client, _ = get_api_client(profile=profile)
    return show_service_token_data(client, token_id)


def create_service_token(
    identifier: str | None = None,
    scopes: list[str] | None = None,
    workspace_ids: list[str] | None = None,
    expires_in_seconds: int | None = None,
    profile: str | None = None,
) -> dict[str, Any]:
    """Create a new service token."""
    client, _ = get_api_client(profile=profile)
    return create_service_token_data(
        client,
        identifier=identifier,
        scopes=scopes,
        workspace_ids=workspace_ids,
        expires_in_seconds=expires_in_seconds,
    )


def rotate_service_token(
    token_id: str,
    overlap_seconds: int = 300,
    expires_in_seconds: int | None = None,
    profile: str | None = None,
) -> dict[str, Any]:
    """Rotate a service token, generating a new secret."""
    client, _ = get_api_client(profile=profile)
    return rotate_service_token_data(
        client,
        token_id,
        overlap_seconds=overlap_seconds,
        expires_in_seconds=expires_in_seconds,
    )


def revoke_service_token(
    token_id: str,
    reason: str,
    profile: str | None = None,
) -> dict[str, str]:
    """Revoke a service token immediately."""
    client, _ = get_api_client(profile=profile)
    return revoke_service_token_data(client, token_id, reason)


__all__ = [
    "create_service_token",
    "list_service_tokens",
    "revoke_service_token",
    "rotate_service_token",
    "show_service_token",
]



"""Bootstrap service token context tests."""

from __future__ import annotations
import asyncio
from datetime import UTC, datetime, timedelta
import pytest
from orcheo_backend.app.authentication import (
    get_authenticator,
    reset_authentication_state,
)
from tests.backend.authentication_test_utils import reset_auth_state


@pytest.fixture(autouse=True)
def _reset_auth(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure authentication state is cleared between tests."""

    yield from reset_auth_state(monkeypatch)


def test_bootstrap_token_grants_default_admin_scopes(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token has default admin scopes."""

    bootstrap_token = "bootstrap-secret-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    reset_authentication_state()

    authenticator = get_authenticator()
    context = None

    async def _authenticate() -> None:
        nonlocal context
        context = await authenticator.authenticate(bootstrap_token)

    asyncio.run(_authenticate())

    assert context is not None
    assert context.subject == "bootstrap"
    assert context.identity_type == "service"
    assert context.token_id == "bootstrap"
    assert "admin:tokens:read" in context.scopes
    assert "admin:tokens:write" in context.scopes
    assert "workflows:read" in context.scopes
    assert "workflows:write" in context.scopes
    assert "workflows:execute" in context.scopes
    assert "vault:read" in context.scopes
    assert "vault:write" in context.scopes


def test_bootstrap_token_respects_custom_scopes(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token can have custom scopes configured."""

    bootstrap_token = "bootstrap-secret-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    monkeypatch.setenv(
        "ORCHEO_AUTH_BOOTSTRAP_TOKEN_SCOPES", "workflows:read,workflows:write"
    )
    reset_authentication_state()

    authenticator = get_authenticator()
    context = None

    async def _authenticate() -> None:
        nonlocal context
        context = await authenticator.authenticate(bootstrap_token)

    asyncio.run(_authenticate())

    assert context is not None
    assert context.scopes == frozenset(["workflows:read", "workflows:write"])
    assert "admin:tokens:write" not in context.scopes


def test_bootstrap_token_has_no_workspace_restrictions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token has no workspace restrictions."""

    bootstrap_token = "bootstrap-secret-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    reset_authentication_state()

    authenticator = get_authenticator()
    context = None

    async def _authenticate() -> None:
        nonlocal context
        context = await authenticator.authenticate(bootstrap_token)

    asyncio.run(_authenticate())

    assert context is not None
    assert context.workspace_ids == frozenset()


def test_bootstrap_token_defaults_to_no_expiration(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token does not expire when no expiry is configured."""

    bootstrap_token = "bootstrap-secret-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    reset_authentication_state()

    authenticator = get_authenticator()
    context = None

    async def _authenticate() -> None:
        nonlocal context
        context = await authenticator.authenticate(bootstrap_token)

    asyncio.run(_authenticate())

    assert context is not None
    assert context.expires_at is None
    assert context.issued_at is None


def test_bootstrap_token_honours_future_expiration(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token returns configured expiry when still valid."""

    bootstrap_token = "bootstrap-secret-token"
    expires_at = datetime.now(tz=UTC) + timedelta(minutes=10)
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_TOKEN_EXPIRES_AT", expires_at.isoformat())
    reset_authentication_state()

    authenticator = get_authenticator()
    context = None

    async def _authenticate() -> None:
        nonlocal context
        context = await authenticator.authenticate(bootstrap_token)

    asyncio.run(_authenticate())

    assert context is not None
    assert context.expires_at == expires_at
    assert context.claims.get("expires_at") == expires_at.isoformat()



"""Workflow tool behavior tests for AI nodes."""

from __future__ import annotations
from unittest.mock import patch
import pytest
from langchain_core.runnables import RunnableConfig
from pydantic import BaseModel
from orcheo.graph.state import State
from orcheo.nodes.ai import WorkflowTool, _create_workflow_tool_func


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_with_workflow_tools(
    mock_mcp_client_class, mock_create_agent, agent, mock_agent, mock_mcp_client
):
    """Workflow tools should be included when provided."""

    from langgraph.graph import StateGraph

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    workflow_graph = StateGraph(dict)

    def simple_node(state):
        return {"result": "workflow result"}

    workflow_graph.add_node("start", simple_node)
    workflow_graph.set_entry_point("start")
    workflow_graph.set_finish_point("start")

    class WorkflowArgs(BaseModel):
        input_value: str

    workflow_tool = WorkflowTool(
        name="test_workflow",
        description="A test workflow tool",
        graph=workflow_graph,
        args_schema=WorkflowArgs,
    )

    agent.predefined_tools = []
    agent.workflow_tools = [workflow_tool]
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 1
    assert call_kwargs["tools"][0].name == "test_workflow"


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_with_workflow_tools_no_args_schema(
    mock_mcp_client_class, mock_create_agent, agent, mock_agent, mock_mcp_client
):
    """Workflow tools without schemas should still be wired up."""

    from langgraph.graph import StateGraph

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    workflow_graph = StateGraph(dict)

    def simple_node(state):
        return {"result": "workflow result"}

    workflow_graph.add_node("start", simple_node)
    workflow_graph.set_entry_point("start")
    workflow_graph.set_finish_point("start")

    workflow_tool = WorkflowTool(
        name="test_workflow_no_schema",
        description="A test workflow tool without schema",
        graph=workflow_graph,
        args_schema=None,
    )

    agent.predefined_tools = []
    agent.workflow_tools = [workflow_tool]
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 1
    assert call_kwargs["tools"][0].name == "test_workflow_no_schema"


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_workflow_tool_compilation_caching(
    mock_mcp_client_class, mock_create_agent, agent, mock_agent, mock_mcp_client
):
    """WorkflowTool should compile graphs once and reuse them."""

    from langgraph.graph import StateGraph

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    workflow_graph = StateGraph(dict)

    def simple_node(state):
        return {"result": "workflow result"}

    workflow_graph.add_node("start", simple_node)
    workflow_graph.set_entry_point("start")
    workflow_graph.set_finish_point("start")

    workflow_tool = WorkflowTool(
        name="cached_workflow",
        description="A workflow with cached compilation",
        graph=workflow_graph,
    )

    compiled_1 = workflow_tool.get_compiled_graph()
    compiled_2 = workflow_tool.get_compiled_graph()
    assert compiled_1 is compiled_2

    agent.predefined_tools = []
    agent.workflow_tools = [workflow_tool]
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 1


@pytest.mark.asyncio
async def test_workflow_tool_async_execution():
    """Async execution path should invoke compiled graph via ainvoke."""

    from langgraph.graph import StateGraph

    workflow_graph = StateGraph(dict)
    execution_tracker = {"executed": False}

    def test_node(state):
        execution_tracker["executed"] = True
        return {"result": "async_executed", "input_data": state}

    workflow_graph.add_node("process", test_node)
    workflow_graph.set_entry_point("process")
    workflow_graph.set_finish_point("process")

    compiled_graph = workflow_graph.compile()

    tool = _create_workflow_tool_func(
        compiled_graph=compiled_graph,
        name="test_async_tool",
        description="Test async execution",
        args_schema=None,
    )

    result = await tool.ainvoke({"test_input": "test_value"})

    assert execution_tracker["executed"]
    assert result["result"] == "async_executed"


def test_workflow_tool_sync_execution():
    """Sync execution path should funnel through asyncio.run."""

    from langgraph.graph import StateGraph

    workflow_graph = StateGraph(dict)
    execution_tracker = {"executed": False}

    def test_node(state):
        execution_tracker["executed"] = True
        return {"result": "sync_executed", "input_data": state}

    workflow_graph.add_node("process", test_node)
    workflow_graph.set_entry_point("process")
    workflow_graph.set_finish_point("process")

    compiled_graph = workflow_graph.compile()

    tool = _create_workflow_tool_func(
        compiled_graph=compiled_graph,
        name="test_sync_tool",
        description="Test sync execution",
        args_schema=None,
    )

    result = tool.invoke({"test_input": "sync_test"})

    assert execution_tracker["executed"]
    assert result["result"] == "sync_executed"



"""Tests for MCP server node discovery endpoints."""

from __future__ import annotations
import pytest
from orcheo_sdk.cli.errors import CLIError


def test_list_nodes(mock_env: None) -> None:
    """Test listing nodes."""
    from orcheo_sdk.mcp_server import tools

    result = tools.list_nodes()
    assert isinstance(result, list)
    assert len(result) > 0
    node_names = [node["name"] for node in result]
    assert "WebhookTriggerNode" in node_names


def test_list_nodes_with_tag_filter(mock_env: None) -> None:
    """Test listing nodes with tag filter."""
    from orcheo_sdk.mcp_server import tools

    result = tools.list_nodes(tag="trigger")
    assert isinstance(result, list)
    for node in result:
        assert (
            "trigger" in node["name"].lower() or "trigger" in node["category"].lower()
        )


def test_show_node_success(mock_env: None) -> None:
    """Test showing node details."""
    from orcheo_sdk.mcp_server import tools

    result = tools.show_node("WebhookTriggerNode")
    assert result["name"] == "WebhookTriggerNode"
    assert "category" in result
    assert "description" in result
    assert "schema" in result


def test_show_node_not_found(mock_env: None) -> None:
    """Test showing non-existent node."""
    from orcheo_sdk.mcp_server import tools

    with pytest.raises(CLIError, match="not registered"):
        tools.show_node("NonExistentNode")


def test_show_node_with_attributes_only(mock_env: None) -> None:
    """Test showing node that has attributes but no model_json_schema."""
    from orcheo.nodes.registry import NodeMetadata, registry

    test_meta = NodeMetadata(
        name="TestNodeWithAttrs",
        description="Test node with attributes",
        category="test",
    )

    class TestNodeWithAttrs:
        """Node with annotations but no model_json_schema."""

        test_attr: str
        count: int

    registry._nodes["TestNodeWithAttrs"] = TestNodeWithAttrs
    registry._metadata["TestNodeWithAttrs"] = test_meta

    try:
        from orcheo_sdk.mcp_server import tools

        result = tools.show_node("TestNodeWithAttrs")
        assert result["name"] == "TestNodeWithAttrs"
        assert "attributes" in result
        assert "test_attr" in result["attributes"]
        assert "count" in result["attributes"]
    finally:
        registry._nodes.pop("TestNodeWithAttrs", None)
        registry._metadata.pop("TestNodeWithAttrs", None)


def test_mcp_list_nodes(mock_env: None) -> None:
    """Test list_nodes MCP tool wrapper to cover return statement."""
    import orcheo_sdk.mcp_server.main as main_module

    result = main_module.list_nodes.fn()
    assert isinstance(result, list)


def test_mcp_show_node(mock_env: None) -> None:
    """Test show_node MCP tool wrapper to cover return statement."""
    import orcheo_sdk.mcp_server.main as main_module

    result = main_module.show_node.fn("WebhookTriggerNode")
    assert result["name"] == "WebhookTriggerNode"



"""Base node implementation for Orcheo."""

import logging
from abc import abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any
from langchain_core.runnables import RunnableConfig
from pydantic import BaseModel
from orcheo.graph.state import State
from orcheo.runtime.credentials import (
    CredentialReference,
    CredentialResolverUnavailableError,
    get_active_credential_resolver,
    parse_credential_reference,
)


logger = logging.getLogger(__name__)


class BaseRunnable(BaseModel):
    """Base class for all runnables in Orcheo (nodes and edges).

    Provides common functionality for variable decoding, credential resolution,
    and state management. Does not include tool execution methods, which are
    specific to nodes.
    """

    name: str
    """Unique name of the runnable."""

    def _decode_value(
        self,
        value: Any,
        state: State,
    ) -> Any:
        """Recursively decode a value that may contain template strings."""
        if isinstance(value, CredentialReference):
            return self._resolve_credential_reference(value)
        if isinstance(value, str):
            return self._decode_string_value(value, state)
        if isinstance(value, BaseModel):
            # Handle Pydantic models by decoding their dict representation
            for field_name in value.__class__.model_fields:
                field_value = getattr(value, field_name)
                decoded = self._decode_value(field_value, state)
                setattr(value, field_name, decoded)
            return value
        if isinstance(value, dict):
            return {k: self._decode_value(v, state) for k, v in value.items()}
        if isinstance(value, list):
            return [self._decode_value(item, state) for item in value]
        return value

    def _decode_string_value(self, value: str, state: State) -> Any:
        """Return decoded value for placeholders or state templates."""
        reference = parse_credential_reference(value)
        if reference is not None:
            return self._resolve_credential_reference(reference)
        if "{{" not in value:
            return value

        path_str = value.strip("{}").strip()
        path_parts = path_str.split(".")

        result: Any = state
        for index, part in enumerate(path_parts):
            if isinstance(result, dict) and part in result:
                result = result.get(part)
                continue
            if isinstance(result, BaseModel) and hasattr(result, part):
                result = getattr(result, part)
                continue
            fallback = self._fallback_to_results(path_parts, index, state)
            if fallback is not None:
                result = fallback
                continue
            logger.warning(
                "Runnable %s could not resolve template '%s' at segment '%s'; "
                "leaving value unchanged.",
                self.name,
                value,
                part,
            )
            return value
        return result

    @staticmethod
    def _fallback_to_results(
        path_parts: list[str],
        index: int,
        state: State,
    ) -> Any | None:
        """Return a fallback lookup within ``state['results']`` when applicable."""
        if index != 0 or path_parts[0] == "results":
            return None
        results = state.get("results")
        if not isinstance(results, dict):
            return None
        return results.get(path_parts[index])

    def _resolve_credential_reference(self, reference: CredentialReference) -> Any:
        """Return the materialised value for ``reference`` or raise an error."""
        resolver = get_active_credential_resolver()
        if resolver is None:
            msg = (
                "Credential placeholders require an active resolver. "
                f"Runnable '{self.name}' attempted to access "
                f"{reference.identifier!r}"
            )
            raise CredentialResolverUnavailableError(msg)
        return resolver.resolve(reference)

    def decode_variables(
        self,
        state: State,
        *,
        config: Mapping[str, Any] | None = None,
    ) -> None:
        """Decode the variables in attributes of the runnable."""
        if config is not None and isinstance(state, dict):
            state.setdefault("config", dict(config))
        for key, value in self.__dict__.items():
            self.__dict__[key] = self._decode_value(value, state)


class BaseNode(BaseRunnable):
    """Base class for all nodes in the flow.

    Inherits variable decoding and credential resolution from BaseRunnable,
    and adds tool execution methods specific to nodes.
    """

    def tool_run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the node as a tool."""
        pass  # pragma: no cover

    async def tool_arun(self, *args: Any, **kwargs: Any) -> Any:
        """Async run the node as a tool."""
        pass  # pragma: no cover

    def _serialize_result(self, value: Any) -> Any:
        """Convert Pydantic models inside outputs into serializable primitives."""
        if isinstance(value, BaseModel):
            computed_fields = getattr(
                value.__class__, "__pydantic_computed_fields__", {}
            )
            computed_keys = {
                field.alias or name for name, field in computed_fields.items()
            }
            dumped = value.model_dump()
            for key in computed_keys:  # pragma: no branch
                if key in dumped:
                    dumped.pop(key)
            return self._serialize_result(dumped)
        if isinstance(value, Mapping):
            return {key: self._serialize_result(val) for key, val in value.items()}
        if isinstance(value, tuple):
            return tuple(self._serialize_result(item) for item in value)
        if isinstance(value, Sequence) and not isinstance(
            value, str | bytes | bytearray
        ):
            return [self._serialize_result(item) for item in value]
        return value


class AINode(BaseNode):
    """Base class for all AI nodes in the flow."""

    async def __call__(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Execute the node and wrap the result in a messages key."""
        self.decode_variables(state, config=config)
        result = await self.run(state, config)
        return self._serialize_result(result)

    @abstractmethod
    async def run(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Run the node."""
        pass  # pragma: no cover


class TaskNode(BaseNode):
    """Base class for all non-AI task nodes in the flow."""

    async def __call__(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Execute the node and wrap the result in a outputs key."""
        self.decode_variables(state, config=config)
        result = await self.run(state, config)
        serialized_result = self._serialize_result(result)
        return {"results": {self.name: serialized_result}}

    @abstractmethod
    async def run(
        self, state: State, config: RunnableConfig
    ) -> dict[str, Any] | list[Any]:
        """Run the node."""
        pass  # pragma: no cover


__all__ = ["BaseRunnable", "BaseNode", "AINode", "TaskNode"]



"""Tests for credential health endpoints and helpers."""

from __future__ import annotations
from datetime import UTC, datetime
from uuid import UUID, uuid4
import pytest
from fastapi import HTTPException, Request, status
from starlette.types import Message
from orcheo.models import CredentialHealthStatus
from orcheo.triggers.manual import ManualDispatchItem, ManualDispatchRequest
from orcheo.vault.oauth import (
    CredentialHealthError,
    CredentialHealthReport,
    CredentialHealthResult,
)
from orcheo_backend.app import (
    _credential_service_ref,
    dispatch_cron_triggers,
    dispatch_manual_runs,
    get_workflow_credential_health,
    invoke_webhook_trigger,
    validate_workflow_credentials,
)
from orcheo_backend.app.repository import WorkflowNotFoundError
from orcheo_backend.app.routers import triggers as triggers_router
from orcheo_backend.app.schemas.credentials import CredentialValidationRequest


class _MissingWorkflowRepository:
    async def get_workflow(self, workflow_id):  # pragma: no cover - signature typing
        raise WorkflowNotFoundError("missing")


@pytest.mark.asyncio()
async def test_get_workflow_credential_health_handles_missing_workflow(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """The credential health endpoint raises a 404 for unknown workflows."""

    monkeypatch.setitem(_credential_service_ref, "service", None)

    with pytest.raises(HTTPException) as exc_info:
        await get_workflow_credential_health(
            uuid4(),
            repository=_MissingWorkflowRepository(),
            service=None,
        )

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio()
async def test_get_workflow_credential_health_returns_unknown_response() -> None:
    """A missing cached report results in an UNKNOWN response payload."""

    class Repository:
        async def get_workflow(self, workflow_id):  # noqa: D401 - simple stub
            return object()

    class Service:
        def get_report(self, workflow_id):
            return None

    response = await get_workflow_credential_health(
        uuid4(), repository=Repository(), service=Service()
    )

    assert response.status is CredentialHealthStatus.UNKNOWN
    assert response.credentials == []


@pytest.mark.asyncio()
async def test_get_workflow_credential_health_requires_service() -> None:
    class Repository:
        async def get_workflow(self, workflow_id):
            return object()

    with pytest.raises(HTTPException) as exc_info:
        await get_workflow_credential_health(
            uuid4(), repository=Repository(), service=None
        )

    assert exc_info.value.status_code == status.HTTP_503_SERVICE_UNAVAILABLE


@pytest.mark.asyncio()
async def test_validate_workflow_credentials_reports_failures() -> None:
    workflow_id = uuid4()

    class Repository:
        async def get_workflow(self, workflow_id):
            return object()

    class Service:
        async def ensure_workflow_health(self, workflow_id, *, actor=None):
            report = CredentialHealthReport(
                workflow_id=workflow_id,
                results=[
                    CredentialHealthResult(
                        credential_id=uuid4(),
                        name="Slack",
                        provider="slack",
                        status=CredentialHealthStatus.UNHEALTHY,
                        last_checked_at=datetime.now(tz=UTC),
                        failure_reason="expired",
                    )
                ],
                checked_at=datetime.now(tz=UTC),
            )
            return report

    request = CredentialValidationRequest(actor="ops")
    with pytest.raises(HTTPException) as exc_info:
        await validate_workflow_credentials(
            workflow_id,
            request,
            repository=Repository(),
            service=Service(),
        )

    assert exc_info.value.status_code == status.HTTP_422_UNPROCESSABLE_CONTENT
    assert exc_info.value.detail["failures"] == ["expired"]


@pytest.mark.asyncio()
async def test_validate_workflow_credentials_handles_missing_workflow() -> None:
    request = CredentialValidationRequest(actor="ops")

    with pytest.raises(HTTPException) as exc_info:
        await validate_workflow_credentials(
            uuid4(),
            request,
            repository=_MissingWorkflowRepository(),
            service=None,
        )

    assert exc_info.value.status_code == status.HTTP_404_NOT_FOUND


def _health_error(workflow_id: UUID) -> CredentialHealthError:
    report = CredentialHealthReport(
        workflow_id=workflow_id,
        results=[
            CredentialHealthResult(
                credential_id=uuid4(),
                name="Slack",
                provider="slack",
                status=CredentialHealthStatus.UNHEALTHY,
                last_checked_at=datetime.now(tz=UTC),
                failure_reason="expired",
            )
        ],
        checked_at=datetime.now(tz=UTC),
    )
    return CredentialHealthError(report)


@pytest.mark.asyncio()
async def test_validate_workflow_credentials_requires_service() -> None:
    workflow_id = uuid4()

    class Repository:
        async def get_workflow(self, workflow_id):
            return object()

    request = CredentialValidationRequest(actor="ops")
    with pytest.raises(HTTPException) as exc_info:
        await validate_workflow_credentials(
            workflow_id,
            request,
            repository=Repository(),
            service=None,
        )

    assert exc_info.value.status_code == status.HTTP_503_SERVICE_UNAVAILABLE


@pytest.mark.asyncio()
async def test_invoke_webhook_trigger_wraps_health_error(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workflow_id = uuid4()

    class Repository:
        async def get_latest_version(self, workflow_id):
            return object()

        async def handle_webhook_trigger(self, *args, **kwargs):
            raise _health_error(workflow_id)

    async def _fake_immediate_response(*_args, **_kwargs) -> tuple[None, bool]:
        return None, True

    monkeypatch.setattr(
        triggers_router,
        "_try_immediate_response",
        _fake_immediate_response,
    )

    scope = {
        "type": "http",
        "method": "POST",
        "path": "/",
        "headers": [],
        "query_string": b"",
        "client": ("127.0.0.1", 12345),
    }

    async def receive() -> Message:
        return {"type": "http.request", "body": b"", "more_body": False}

    request = Request(scope, receive)

    with pytest.raises(HTTPException) as exc_info:
        await invoke_webhook_trigger(
            workflow_id,
            request,
            repository=Repository(),
            vault=object(),
        )

    assert exc_info.value.status_code == status.HTTP_422_UNPROCESSABLE_CONTENT


@pytest.mark.asyncio()
async def test_dispatch_cron_triggers_wraps_health_error() -> None:
    workflow_id = uuid4()

    class Repository:
        async def dispatch_due_cron_runs(self, now=None):
            raise _health_error(workflow_id)

    with pytest.raises(HTTPException) as exc_info:
        await dispatch_cron_triggers(repository=Repository())

    assert exc_info.value.status_code == status.HTTP_422_UNPROCESSABLE_CONTENT


@pytest.mark.asyncio()
async def test_dispatch_manual_runs_wraps_health_error() -> None:
    workflow_id = uuid4()

    class Repository:
        async def dispatch_manual_runs(self, request):
            raise _health_error(workflow_id)

    manual_request = ManualDispatchRequest(
        workflow_id=workflow_id,
        actor="ops",
        runs=[ManualDispatchItem(input_payload={})],
    )

    with pytest.raises(HTTPException) as exc_info:
        await dispatch_manual_runs(manual_request, repository=Repository())

    assert exc_info.value.status_code == status.HTTP_422_UNPROCESSABLE_CONTENT



"""OAuth 2.0 PKCE flow implementation for CLI."""

from __future__ import annotations
import base64
import hashlib
import http.server
import json
import secrets
import socketserver
import time
import urllib.parse
import webbrowser
from dataclasses import dataclass
from typing import Any
import httpx
from rich.console import Console
from orcheo_sdk.cli.auth.config import OAuthConfig, get_oauth_config
from orcheo_sdk.cli.auth.tokens import AuthTokens, clear_oauth_tokens, set_oauth_tokens
from orcheo_sdk.cli.errors import CLIError


# Constants matching Canvas implementation
STATE_BYTES = 32
VERIFIER_BYTES = 64
AUTH_STATE_TTL_SECONDS = 600  # 10 minutes
DEFAULT_CALLBACK_PORT = 8085


@dataclass(slots=True)
class OidcDiscovery:
    """OIDC discovery document endpoints."""

    authorization_endpoint: str
    token_endpoint: str
    end_session_endpoint: str | None = None


def _base64_url_encode(data: bytes) -> str:
    """Encode bytes to base64url without padding."""
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode("ascii")


def _create_random_string(length: int) -> str:
    """Generate a cryptographically secure random string."""
    return _base64_url_encode(secrets.token_bytes(length))


def _create_code_challenge(verifier: str) -> str:
    """Create S256 code challenge from verifier."""
    digest = hashlib.sha256(verifier.encode("ascii")).digest()
    return _base64_url_encode(digest)


def _load_discovery(issuer: str) -> OidcDiscovery:
    """Fetch OIDC discovery document."""
    url = f"{issuer.rstrip('/')}/.well-known/openid-configuration"
    try:
        response = httpx.get(url, timeout=30.0)
        response.raise_for_status()
        data = response.json()
    except httpx.HTTPError as exc:
        raise CLIError(f"Failed to load OAuth discovery: {exc}") from exc

    return OidcDiscovery(
        authorization_endpoint=data["authorization_endpoint"],
        token_endpoint=data["token_endpoint"],
        end_session_endpoint=data.get("end_session_endpoint"),
    )


def _parse_jwt_expiry(token: str) -> int | None:
    """Extract expiry timestamp from JWT without validation."""
    try:
        parts = token.split(".")
        if len(parts) < 2:
            return None
        # Add padding if needed
        payload_b64 = parts[1]
        padding_needed = 4 - len(payload_b64) % 4
        if padding_needed != 4:
            payload_b64 += "=" * padding_needed
        payload = json.loads(base64.urlsafe_b64decode(payload_b64))
        exp = payload.get("exp")
        if isinstance(exp, int | float):
            return int(exp * 1000)  # Convert to milliseconds
    except Exception:
        pass
    return None


class _CallbackHandler(http.server.BaseHTTPRequestHandler):
    """HTTP handler for OAuth callback."""

    authorization_code: str | None = None
    callback_state: str | None = None
    error: str | None = None

    def log_message(self, format: str, *args: Any) -> None:
        """Suppress HTTP server logging."""

    def do_GET(self) -> None:  # noqa: N802
        """Handle OAuth callback GET request."""
        parsed = urllib.parse.urlparse(self.path)
        params = urllib.parse.parse_qs(parsed.query)

        if "error" in params:
            _CallbackHandler.error = params["error"][0]
            self._send_response("Authentication failed. You can close this window.")
            return

        if "code" in params and "state" in params:
            _CallbackHandler.authorization_code = params["code"][0]
            _CallbackHandler.callback_state = params["state"][0]
            self._send_response("Authentication successful! You can close this window.")
        else:
            self._send_response("Invalid callback. Missing code or state.")

    def _send_response(self, message: str) -> None:
        """Send HTML response to browser."""
        html = f"""
        <!DOCTYPE html>
        <html>
        <head><title>Orcheo CLI</title></head>
        <body style="font-family: sans-serif; text-align: center; padding: 50px;">
            <h1>Orcheo CLI</h1>
            <p>{message}</p>
        </body>
        </html>
        """
        self.send_response(200)
        self.send_header("Content-Type", "text/html")
        self.end_headers()
        self.wfile.write(html.encode("utf-8"))


def start_oauth_login(
    *,
    console: Console,
    profile: str | None,
    no_browser: bool = False,
    port: int = DEFAULT_CALLBACK_PORT,
) -> None:
    """Execute browser-based OAuth login flow."""
    config = get_oauth_config(profile=profile)
    discovery = _load_discovery(config.issuer)

    # Generate PKCE parameters
    state = _create_random_string(STATE_BYTES)
    verifier = _create_random_string(VERIFIER_BYTES)
    challenge = _create_code_challenge(verifier)

    # Build authorization URL
    redirect_uri = f"http://localhost:{port}/callback"
    auth_params: dict[str, str] = {
        "response_type": "code",
        "client_id": config.client_id,
        "redirect_uri": redirect_uri,
        "scope": config.scopes,
        "state": state,
        "code_challenge": challenge,
        "code_challenge_method": "S256",
    }
    if config.audience:  # pragma: no branch
        auth_params["audience"] = config.audience
    if config.organization:  # pragma: no branch
        auth_params["organization"] = config.organization

    auth_url = (
        f"{discovery.authorization_endpoint}?{urllib.parse.urlencode(auth_params)}"
    )

    # Reset handler state
    _CallbackHandler.authorization_code = None
    _CallbackHandler.callback_state = None
    _CallbackHandler.error = None

    # Start callback server
    with socketserver.TCPServer(("", port), _CallbackHandler) as server:
        server.timeout = AUTH_STATE_TTL_SECONDS

        if no_browser:
            console.print(
                f"\nOpen this URL in your browser:\n[cyan]{auth_url}[/cyan]\n"
            )
        else:
            console.print("Opening browser for authentication...")
            webbrowser.open(auth_url)

        console.print(f"Waiting for callback on port {port}...")

        # Wait for callback with timeout
        start_time = time.time()
        while (time.time() - start_time) < AUTH_STATE_TTL_SECONDS:
            server.handle_request()
            if _CallbackHandler.authorization_code or _CallbackHandler.error:
                break

        if _CallbackHandler.error:
            raise CLIError(f"OAuth error: {_CallbackHandler.error}")

        if not _CallbackHandler.authorization_code:
            raise CLIError("Authentication timed out. Please try again.")

        if _CallbackHandler.callback_state != state:
            raise CLIError("OAuth state mismatch. Possible CSRF attack.")

        # Exchange code for tokens
        console.print("Exchanging authorization code for tokens...")
        tokens = _exchange_code(
            config=config,
            discovery=discovery,
            code=_CallbackHandler.authorization_code,
            verifier=verifier,
            redirect_uri=redirect_uri,
        )

        set_oauth_tokens(profile=profile, tokens=tokens)
        console.print("[green]Successfully authenticated![/green]")


def _exchange_code(
    *,
    config: OAuthConfig,
    discovery: OidcDiscovery,
    code: str,
    verifier: str,
    redirect_uri: str,
) -> AuthTokens:
    """Exchange authorization code for tokens."""
    body = {
        "grant_type": "authorization_code",
        "client_id": config.client_id,
        "code": code,
        "redirect_uri": redirect_uri,
        "code_verifier": verifier,
    }

    try:
        response = httpx.post(
            discovery.token_endpoint,
            data=body,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            timeout=30.0,
        )
        response.raise_for_status()
        data = response.json()
    except httpx.HTTPError as exc:
        raise CLIError(f"Token exchange failed: {exc}") from exc

    access_token = data.get("access_token")
    if not access_token:
        raise CLIError("Token response missing access_token")

    # Calculate expiry
    expires_at: int | None
    expires_in = data.get("expires_in")
    if isinstance(expires_in, int | float):
        expires_at = int((time.time() + expires_in) * 1000)
    else:
        expires_at = _parse_jwt_expiry(access_token) or _parse_jwt_expiry(
            data.get("id_token", "")
        )

    return AuthTokens(
        access_token=access_token,
        id_token=data.get("id_token"),
        refresh_token=data.get("refresh_token"),
        token_type=data.get("token_type", "Bearer"),
        expires_at=expires_at,
    )


def logout_oauth(*, profile: str | None) -> None:
    """Clear OAuth tokens for the given profile."""
    clear_oauth_tokens(profile=profile)



"""Workflow download CLI tests for auto format logic."""

from __future__ import annotations
import json
import httpx
import respx
from typer.testing import CliRunner
from orcheo_sdk.cli.main import app


def test_workflow_download_auto_format_langgraph_script(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Test download with auto format: LangGraph script downloads as Python."""
    original_source = """from langgraph.graph import StateGraph

def my_node(state):
    return state

graph = StateGraph(dict)
graph.add_node("my_node", my_node)
"""
    workflow = {"id": "wf-1", "name": "LangGraphWorkflow"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            "graph": {
                "format": "langgraph-script",
                "source": original_source,
                "entrypoint": None,
                "summary": {"nodes": [{"name": "my_node"}], "edges": []},
            },
        }
    ]

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        result = runner.invoke(
            app,
            ["workflow", "download", "wf-1"],  # No --format, should default to auto
            env=env,
        )
    assert result.exit_code == 0
    # Should output Python code (not JSON)
    assert "from langgraph.graph import StateGraph" in result.stdout
    assert "def my_node(state)" in result.stdout
    # Should NOT contain JSON structure
    assert '"name"' not in result.stdout


def test_workflow_download_auto_format_sdk_workflow(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Test workflow download with auto format for SDK workflow downloads as JSON."""
    workflow = {"id": "wf-1", "name": "SDKWorkflow"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            "graph": {
                # No "format" field or format is not "langgraph-script"
                "nodes": [{"name": "node1", "type": "Agent"}],
                "edges": [],
            },
        }
    ]

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        result = runner.invoke(
            app,
            ["workflow", "download", "wf-1"],  # No --format, should default to auto
            env=env,
        )
    assert result.exit_code == 0
    # Should output JSON (not Python code)
    output = json.loads(result.stdout)
    assert output["name"] == "SDKWorkflow"
    assert "graph" in output


def test_workflow_download_auto_format_missing_graph(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Gracefully handle workflows whose versions lack a graph payload."""
    workflow = {"id": "wf-1", "name": "SDKWorkflow"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            # Older versions may omit the "graph" field altogether.
            # The CLI should fall back to JSON output without crashing.
        }
    ]

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        result = runner.invoke(
            app,
            ["workflow", "download", "wf-1"],  # No --format, should default to auto
            env=env,
        )
    assert result.exit_code == 0
    output = json.loads(result.stdout)
    assert output["name"] == "SDKWorkflow"
    assert output["graph"] == {}


def test_workflow_download_auto_format_explicit(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Test workflow download with explicit --format auto for LangGraph script."""
    original_source = "from langgraph.graph import StateGraph\n"
    workflow = {"id": "wf-1", "name": "Test"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            "graph": {
                "format": "langgraph-script",
                "source": original_source,
                "entrypoint": None,
                "summary": {"nodes": [], "edges": []},
            },
        }
    ]

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        result = runner.invoke(
            app,
            ["workflow", "download", "wf-1", "--format", "auto"],
            env=env,
        )
    assert result.exit_code == 0
    # Should output Python code
    assert original_source in result.stdout


def test_workflow_download_with_cache_notice(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Test workflow download shows cache notice when using cached data."""
    workflow = {"id": "wf-1", "name": "Test"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            "graph": {"nodes": [{"id": "a"}], "edges": []},
        }
    ]

    # First call to populate cache
    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        first = runner.invoke(
            app,
            ["workflow", "download", "wf-1"],
            env=env,
        )
        assert first.exit_code == 0

    # Second call in offline mode should use cache and show notice
    result = runner.invoke(
        app,
        ["--offline", "workflow", "download", "wf-1"],
        env=env,
    )
    assert result.exit_code == 0
    assert "Using cached data" in result.stdout



"""Tests for workflow CLI input resolution helpers."""

from __future__ import annotations
from pathlib import Path
from unittest.mock import MagicMock, call
import pytest
from orcheo_sdk.cli.errors import CLIError
from orcheo_sdk.cli.state import CLIState
from orcheo_sdk.cli.workflow.inputs import (
    _cache_notice,
    _load_inputs_from_path,
    _load_inputs_from_string,
    _resolve_evaluation_payload,
    _resolve_run_inputs,
    _resolve_runnable_config,
    _validate_local_path,
)


def test_resolve_run_inputs_from_string() -> None:
    payload = '{"key": "value"}'
    resolved = _resolve_run_inputs(payload, None)
    assert resolved == {"key": "value"}


def test_resolve_run_inputs_from_file(tmp_path: Path) -> None:
    inputs_file = tmp_path / "inputs.json"
    inputs_file.write_text('{"flag": true}', encoding="utf-8")

    resolved = _resolve_run_inputs(None, str(inputs_file))
    assert resolved == {"flag": True}


def test_resolve_run_inputs_conflict() -> None:
    with pytest.raises(CLIError, match="either --inputs or --inputs-file"):
        _resolve_run_inputs("{}", "/tmp/inputs.json")


def test_resolve_runnable_config_from_string() -> None:
    config = '{"mode": "fast"}'
    resolved = _resolve_runnable_config(config, None)
    assert resolved == {"mode": "fast"}


def test_resolve_runnable_config_from_file(tmp_path: Path) -> None:
    config_file = tmp_path / "config.json"
    config_file.write_text('{"retry": 3}', encoding="utf-8")

    resolved = _resolve_runnable_config(None, str(config_file))
    assert resolved == {"retry": 3}


def test_resolve_runnable_config_conflict(tmp_path: Path) -> None:
    config_file = tmp_path / "config.json"
    config_file.write_text('{"retry": 1}', encoding="utf-8")

    with pytest.raises(CLIError, match="either --config or --config-file"):
        _resolve_runnable_config("{}", str(config_file))


def test_resolve_runnable_config_invalid_payload() -> None:
    with pytest.raises(CLIError, match="Runnable config payload must be a JSON object"):
        _resolve_runnable_config("[]", None)


def test_resolve_runnable_config_returns_none_when_missing() -> None:
    assert _resolve_runnable_config(None, None) is None


def test_resolve_evaluation_payload_requires_value() -> None:
    with pytest.raises(CLIError, match="Provide --evaluation or --evaluation-file"):
        _resolve_evaluation_payload(None, None)


def test_resolve_evaluation_payload_conflict(tmp_path: Path) -> None:
    eval_file = tmp_path / "evaluation.json"
    eval_file.write_text('{"dataset": {"cases": [{"inputs": {}}]}}', encoding="utf-8")

    with pytest.raises(CLIError, match="either --evaluation or --evaluation-file"):
        _resolve_evaluation_payload("{}", str(eval_file))


def test_resolve_evaluation_payload_from_file(tmp_path: Path) -> None:
    eval_file = tmp_path / "evaluation.json"
    eval_file.write_text('{"dataset": {"cases": [{"inputs": {}}]}}', encoding="utf-8")

    resolved = _resolve_evaluation_payload(None, str(eval_file))
    assert resolved["dataset"]["cases"][0]["inputs"] == {}


def test_resolve_evaluation_payload_from_string() -> None:
    payload = '{"dataset": {"cases": [{"inputs": {"value": 1}}]}}'
    resolved = _resolve_evaluation_payload(payload, None)
    assert resolved["dataset"]["cases"][0]["inputs"]["value"] == 1


def test_load_inputs_from_string_requires_mapping() -> None:
    with pytest.raises(CLIError, match="Inputs payload must be a JSON object"):
        _load_inputs_from_string("[]")


def test_load_inputs_from_path_requires_mapping(tmp_path: Path) -> None:
    inputs_file = tmp_path / "inputs.json"
    inputs_file.write_text("[]", encoding="utf-8")

    with pytest.raises(CLIError, match="Inputs payload must be a JSON object"):
        _load_inputs_from_path(str(inputs_file))


def test_validate_local_path_rejects_relative_outside() -> None:
    with pytest.raises(CLIError, match="escapes the current working directory"):
        _validate_local_path("..", description="inputs")


def test_validate_local_path_requires_existing_file(tmp_path: Path) -> None:
    missing = tmp_path / "missing.json"

    with pytest.raises(CLIError, match="does not exist"):
        _validate_local_path(str(missing), description="inputs")


def test_validate_local_path_requires_file(tmp_path: Path) -> None:
    directory = tmp_path / "folder"
    directory.mkdir()

    with pytest.raises(CLIError, match="is not a file"):
        _validate_local_path(str(directory), description="inputs")


def test_validate_local_path_disallows_directory_when_not_required(
    tmp_path: Path,
) -> None:
    directory = tmp_path / "folder"
    directory.mkdir()

    with pytest.raises(CLIError, match="is not a file"):
        _validate_local_path(
            str(directory), description="inputs", must_exist=False, require_file=True
        )


def test_validate_local_path_requires_parent_directory(tmp_path: Path) -> None:
    target_path = tmp_path / "missing" / "inputs.json"

    with pytest.raises(CLIError, match="does not exist"):
        _validate_local_path(
            str(target_path),
            description="config",
            must_exist=False,
        )


def test_validate_local_path_parent_must_be_directory(tmp_path: Path) -> None:
    parent_file = tmp_path / "parent-file"
    parent_file.write_text("", encoding="utf-8")

    child_path = parent_file / "child.json"

    with pytest.raises(CLIError, match="Parent of config path"):
        _validate_local_path(
            str(child_path),
            description="config",
            must_exist=False,
        )


def test_cache_notice_reports_staleness_flags() -> None:
    console = MagicMock()
    state = CLIState(
        settings=MagicMock(),
        client=MagicMock(),
        cache=MagicMock(),
        console=console,
        human=True,
    )

    _cache_notice(state, "workflow run", stale=False)
    _cache_notice(state, "evaluation", stale=True)
    console.print.assert_has_calls(
        [
            call("[yellow]Using cached data[/yellow] for workflow run."),
            call("[yellow]Using cached data[/yellow] (older than TTL) for evaluation."),
        ]
    )



"""Scope enforcement tests for InMemoryCredentialVault."""

from __future__ import annotations
from uuid import UUID, uuid4
import pytest
from orcheo.models import (
    AesGcmCredentialCipher,
    CredentialAccessContext,
    CredentialHealthStatus,
    CredentialKind,
    CredentialScope,
)
from orcheo.vault import (
    CredentialNotFoundError,
    InMemoryCredentialVault,
    RotationPolicyError,
    WorkflowScopeError,
)


def test_inmemory_vault_supports_shared_and_restricted_credentials() -> None:
    cipher = AesGcmCredentialCipher(key="unit-test-key")
    vault = InMemoryCredentialVault(cipher=cipher)
    workflow_a, workflow_b = uuid4(), uuid4()
    context_a, context_b = (
        CredentialAccessContext(workflow_id=workflow_a),
        CredentialAccessContext(workflow_id=workflow_b),
    )

    metadata = vault.create_credential(
        name="OpenAI",
        provider="openai",
        scopes=["chat:write"],
        secret="initial-token",
        actor="alice",
    )

    assert metadata.kind is CredentialKind.SECRET
    assert metadata.health.status is CredentialHealthStatus.UNKNOWN

    assert (
        vault.reveal_secret(credential_id=metadata.id, context=context_a)
        == "initial-token"
    )
    assert (
        vault.reveal_secret(credential_id=metadata.id, context=context_b)
        == "initial-token"
    )

    assert [item.id for item in vault.list_credentials(context=context_a)] == [
        metadata.id
    ]
    assert [item.id for item in vault.list_credentials(context=context_b)] == [
        metadata.id
    ]

    masked = vault.describe_credentials(context=context_a)
    assert masked[0]["encryption"]["algorithm"] == cipher.algorithm
    assert "ciphertext" not in masked[0]["encryption"]
    assert masked[0]["scope"]["workflow_ids"] == []
    assert masked[0]["kind"] == "secret"
    assert masked[0]["health"]["status"] == CredentialHealthStatus.UNKNOWN.value

    with pytest.raises(RotationPolicyError):
        vault.rotate_secret(
            credential_id=metadata.id,
            secret="initial-token",
            actor="security-bot",
            context=context_a,
        )

    rotated = vault.rotate_secret(
        credential_id=metadata.id,
        secret="rotated-token",
        actor="security-bot",
        context=context_a,
    )
    assert rotated.last_rotated_at >= metadata.last_rotated_at
    assert rotated.health.status is CredentialHealthStatus.UNKNOWN
    assert (
        vault.reveal_secret(credential_id=metadata.id, context=context_a)
        == "rotated-token"
    )
    assert (
        vault.reveal_secret(credential_id=metadata.id, context=context_b)
        == "rotated-token"
    )

    shared_entry = vault.describe_credentials(context=context_b)[0]
    assert shared_entry["scope"]["workflow_ids"] == []
    assert shared_entry["health"]["status"] == CredentialHealthStatus.UNKNOWN.value

    restricted_scope = CredentialScope.for_workflows(workflow_a)
    restricted = vault.create_credential(
        name="Slack",
        provider="slack",
        scopes=["chat:write"],
        secret="slack-token",
        actor="alice",
        scope=restricted_scope,
    )

    assert (
        vault.reveal_secret(credential_id=restricted.id, context=context_a)
        == "slack-token"
    )

    with pytest.raises(WorkflowScopeError):
        vault.reveal_secret(credential_id=restricted.id, context=context_b)

    assert {item.id for item in vault.list_credentials(context=context_a)} == {
        metadata.id,
        restricted.id,
    }
    assert [item.id for item in vault.list_credentials(context=context_b)] == [
        metadata.id
    ]

    role_scope = CredentialScope.for_roles("admin")
    role_metadata = vault.create_credential(
        name="PagerDuty",
        provider="pagerduty",
        scopes=[],
        secret="pd-key",
        actor="alice",
        scope=role_scope,
    )

    admin_context, viewer_context = (
        CredentialAccessContext(roles=["Admin", "operator"]),
        CredentialAccessContext(roles=["viewer"]),
    )

    assert (
        vault.reveal_secret(credential_id=role_metadata.id, context=admin_context)
        == "pd-key"
    )

    with pytest.raises(WorkflowScopeError):
        vault.reveal_secret(credential_id=role_metadata.id, context=viewer_context)

    unknown_id = UUID(int=0)
    with pytest.raises(CredentialNotFoundError):
        vault.reveal_secret(credential_id=unknown_id, context=context_a)
    viewer_describe = vault.describe_credentials(context=viewer_context)
    assert [entry["id"] for entry in viewer_describe] == [str(metadata.id)]
    assert viewer_describe[0]["scope"]["roles"] == []


def test_delete_credential_enforces_scope() -> None:
    cipher = AesGcmCredentialCipher(key="delete-scope")
    vault = InMemoryCredentialVault(cipher=cipher)
    workflow_id = uuid4()

    metadata = vault.create_credential(
        name="Restricted",
        provider="service",
        scopes=["read"],
        secret="secret",
        actor="ops",
        scope=CredentialScope.for_workflows(workflow_id),
    )

    with pytest.raises(WorkflowScopeError):
        vault.delete_credential(
            metadata.id,
            context=CredentialAccessContext(workflow_id=uuid4()),
        )



"""SQLite-backed workflow repository split across focused mixins."""

from __future__ import annotations
from pathlib import Path
from orcheo.vault.oauth import OAuthCredentialService
from orcheo_backend.app.repository_sqlite._retry import RetryPolicyMixin
from orcheo_backend.app.repository_sqlite._runs import WorkflowRunMixin
from orcheo_backend.app.repository_sqlite._triggers import TriggerRepositoryMixin
from orcheo_backend.app.repository_sqlite._versions import WorkflowVersionMixin
from orcheo_backend.app.repository_sqlite._workflows import WorkflowRepositoryMixin


class SqliteWorkflowRepository(
    TriggerRepositoryMixin,
    RetryPolicyMixin,
    WorkflowRunMixin,
    WorkflowVersionMixin,
    WorkflowRepositoryMixin,
):
    """SQLite-backed workflow repository for durable local development state."""

    def __init__(
        self,
        database_path: str | Path,
        *,
        credential_service: OAuthCredentialService | None = None,
    ) -> None:
        """Configure repository dependencies and initialize trigger layer."""
        super().__init__(database_path, credential_service=credential_service)


__all__ = ["SqliteWorkflowRepository"]



"""Bootstrap configuration parsing tests."""

from __future__ import annotations
from datetime import UTC, datetime, timedelta
import pytest
from orcheo_backend.app.authentication import (
    _parse_timestamp,
    load_auth_settings,
    reset_authentication_state,
)
from tests.backend.authentication_test_utils import reset_auth_state


@pytest.fixture(autouse=True)
def _reset_auth(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure authentication state is cleared between tests."""

    yield from reset_auth_state(monkeypatch)


def test_bootstrap_token_enforces_authentication_mode(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token enables enforce mode when AUTH_MODE is optional."""

    bootstrap_token = "bootstrap-secret-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    monkeypatch.setenv("ORCHEO_AUTH_MODE", "optional")
    reset_authentication_state()

    settings = load_auth_settings()
    assert settings.enforce is True
    assert settings.bootstrap_service_token == bootstrap_token


def test_load_auth_settings_includes_bootstrap_token(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """load_auth_settings correctly parses bootstrap token configuration."""

    bootstrap_token = "my-bootstrap-token"
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    monkeypatch.setenv(
        "ORCHEO_AUTH_BOOTSTRAP_TOKEN_SCOPES", "workflows:read,vault:write"
    )
    reset_authentication_state()

    settings = load_auth_settings()
    assert settings.bootstrap_service_token == bootstrap_token
    assert settings.bootstrap_token_scopes == frozenset(
        ["workflows:read", "vault:write"]
    )
    assert settings.bootstrap_token_expires_at is None


def test_load_auth_settings_parses_bootstrap_token_expiration(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Bootstrap token expiration is parsed into an aware datetime."""

    bootstrap_token = "my-bootstrap-token"
    expires_at = datetime.now(tz=UTC) + timedelta(hours=1)
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_SERVICE_TOKEN", bootstrap_token)
    monkeypatch.setenv("ORCHEO_AUTH_BOOTSTRAP_TOKEN_EXPIRES_AT", expires_at.isoformat())
    reset_authentication_state()

    settings = load_auth_settings()
    assert settings.bootstrap_service_token == bootstrap_token
    assert settings.bootstrap_token_expires_at == expires_at


def test_parse_timestamp_with_naive_datetime() -> None:
    """_parse_timestamp converts naive datetime to aware datetime."""

    naive_dt = datetime(2024, 11, 4, 12, 0, 0)
    result = _parse_timestamp(naive_dt)

    assert result is not None
    assert result.tzinfo == UTC
    assert result.year == 2024
    assert result.month == 11
    assert result.day == 4
    assert result.hour == 12
    assert result.minute == 0
    assert result.second == 0



from __future__ import annotations
from datetime import UTC, datetime
from uuid import uuid4
import pytest
from orcheo.triggers.cron import CronTriggerConfig
from orcheo.triggers.webhook import WebhookTriggerConfig
from orcheo_backend.app.repository import (
    InMemoryWorkflowRepository,
    WorkflowNotFoundError,
    WorkflowPublishStateError,
    WorkflowVersionNotFoundError,
)


@pytest.mark.asyncio()
async def test_inmemory_latest_version_missing_instance() -> None:
    """Missing latest version objects surface a dedicated error."""

    repository = InMemoryWorkflowRepository()

    workflow = await repository.create_workflow(
        name="Latest", slug=None, description=None, tags=None, actor="tester"
    )
    version = await repository.create_version(
        workflow.id,
        graph={},
        metadata={},
        notes=None,
        created_by="tester",
    )

    repository._versions.pop(version.id)  # noqa: SLF001

    with pytest.raises(WorkflowVersionNotFoundError):
        await repository.get_latest_version(workflow.id)


@pytest.mark.asyncio()
async def test_inmemory_handle_webhook_missing_version_object() -> None:
    """Webhook dispatch raises when the latest version is missing."""

    repository = InMemoryWorkflowRepository()

    workflow = await repository.create_workflow(
        name="Webhook", slug=None, description=None, tags=None, actor="tester"
    )
    version = await repository.create_version(
        workflow.id,
        graph={},
        metadata={},
        notes=None,
        created_by="tester",
    )
    await repository.configure_webhook_trigger(
        workflow.id, WebhookTriggerConfig(allowed_methods={"POST"})
    )

    repository._versions.pop(version.id)  # noqa: SLF001

    with pytest.raises(WorkflowVersionNotFoundError):
        await repository.handle_webhook_trigger(
            workflow.id,
            method="POST",
            headers={},
            query_params={},
            payload={},
            source_ip=None,
        )


@pytest.mark.asyncio()
async def test_inmemory_cron_dispatch_skips_missing_versions() -> None:
    """Cron dispatch ignores schedules when the latest version is missing."""

    repository = InMemoryWorkflowRepository()

    workflow = await repository.create_workflow(
        name="Cron", slug=None, description=None, tags=None, actor="owner"
    )
    version = await repository.create_version(
        workflow.id,
        graph={},
        metadata={},
        notes=None,
        created_by="owner",
    )
    await repository.configure_cron_trigger(
        workflow.id, CronTriggerConfig(expression="0 12 * * *", timezone="UTC")
    )

    repository._versions.pop(version.id)  # noqa: SLF001

    runs = await repository.dispatch_due_cron_runs(
        now=datetime(2025, 1, 1, 12, 0, tzinfo=UTC)
    )
    assert runs == []


@pytest.mark.asyncio()
async def test_inmemory_publish_workflow_missing_id_raises_not_found() -> None:
    """publish_workflow raises WorkflowNotFoundError for unknown IDs."""

    repository = InMemoryWorkflowRepository()

    with pytest.raises(WorkflowNotFoundError):
        await repository.publish_workflow(
            uuid4(),
            require_login=False,
            actor="tester",
        )


@pytest.mark.asyncio()
async def test_inmemory_publish_workflow_translates_value_errors() -> None:
    """ValueError from Workflow.publish is surfaced as WorkflowPublishStateError."""

    repository = InMemoryWorkflowRepository()
    workflow = await repository.create_workflow(
        name="Publish Twice", slug=None, description=None, tags=None, actor="tester"
    )
    await repository.publish_workflow(
        workflow.id,
        require_login=False,
        actor="tester",
    )

    with pytest.raises(WorkflowPublishStateError):
        await repository.publish_workflow(
            workflow.id,
            require_login=False,
            actor="tester",
        )


@pytest.mark.asyncio()
async def test_inmemory_revoke_publish_missing_workflow() -> None:
    """revoke_publish raises WorkflowNotFoundError for unknown workflows."""

    repository = InMemoryWorkflowRepository()

    with pytest.raises(WorkflowNotFoundError):
        await repository.revoke_publish(uuid4(), actor="tester")


@pytest.mark.asyncio()
async def test_inmemory_revoke_publish_requires_published_state() -> None:
    """revoke_publish also translates invalid state errors."""

    repository = InMemoryWorkflowRepository()
    workflow = await repository.create_workflow(
        name="Revoke",
        slug=None,
        description=None,
        tags=None,
        actor="tester",
    )

    with pytest.raises(WorkflowPublishStateError):
        await repository.revoke_publish(workflow.id, actor="tester")



"""Workflow-related request/response schemas."""

from __future__ import annotations
from typing import Any
from uuid import UUID
from pydantic import BaseModel, Field, field_validator
from orcheo.graph.ingestion import DEFAULT_SCRIPT_SIZE_LIMIT
from orcheo.models.workflow import Workflow
from orcheo.runtime.runnable_config import RunnableConfigModel


class WorkflowCreateRequest(BaseModel):
    """Payload for creating a new workflow."""

    name: str
    slug: str | None = None
    description: str | None = None
    tags: list[str] = Field(default_factory=list)
    actor: str = Field(default="system")


class WorkflowUpdateRequest(BaseModel):
    """Payload for updating an existing workflow."""

    name: str | None = None
    description: str | None = None
    tags: list[str] | None = None
    is_archived: bool | None = None
    actor: str = Field(default="system")


class WorkflowVersionCreateRequest(BaseModel):
    """Payload for creating a workflow version."""

    graph: dict[str, Any]
    metadata: dict[str, Any] = Field(default_factory=dict)
    runnable_config: RunnableConfigModel | None = None
    notes: str | None = None
    created_by: str


class WorkflowVersionIngestRequest(BaseModel):
    """Payload for ingesting a LangGraph Python script."""

    script: str
    entrypoint: str | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)
    runnable_config: RunnableConfigModel | None = None
    notes: str | None = None
    created_by: str

    @field_validator("script")
    @classmethod
    def _enforce_script_size(cls, value: str) -> str:
        size = len(value.encode("utf-8"))
        if size > DEFAULT_SCRIPT_SIZE_LIMIT:
            msg = (
                "LangGraph script exceeds the maximum allowed size of "
                f"{DEFAULT_SCRIPT_SIZE_LIMIT} bytes"
            )
            raise ValueError(msg)
        return value


class WorkflowRunCreateRequest(BaseModel):
    """Payload for creating a new workflow execution run."""

    workflow_version_id: UUID
    triggered_by: str
    input_payload: dict[str, Any] = Field(default_factory=dict)
    runnable_config: RunnableConfigModel | None = None


class WorkflowVersionDiffResponse(BaseModel):
    """Response payload for workflow version diffs."""

    base_version: int
    target_version: int
    diff: list[str]


class WorkflowPublishRequest(BaseModel):
    """Payload for publishing a workflow."""

    require_login: bool = False
    actor: str = Field(default="system")


class WorkflowPublishRevokeRequest(BaseModel):
    """Payload for revoking workflow publication."""

    actor: str = Field(default="system")


class WorkflowPublishResponse(BaseModel):
    """Response payload for publish actions."""

    workflow: Workflow
    message: str | None = None
    share_url: str | None = None


class PublicWorkflow(BaseModel):
    """Public workflow metadata returned without authentication."""

    id: UUID
    name: str
    description: str | None = None
    is_public: bool
    require_login: bool
    share_url: str | None = None
